#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#     .. code-block:: python
#
#             from PAMI.fuzzyGeoreferencedFrequentPattern import FFSPMiner as alg
#
#             minSup = str()
#
#             iFile = " "
#
#             sep = "\t"
#
#             obj = alg.FFSPMiner("input.txt", "neighbours.txt", 2)
#
#             obj.mine()
#
#             fuzzySpatialFrequentPatterns = obj.getPatterns()
#
#             print("Total number of fuzzy frequent spatial patterns:", len(fuzzySpatialFrequentPatterns))
#
#             obj.save("outputFile")
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#



__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.fuzzyGeoreferencedFrequentPattern.basic import abstract as _ab
from deprecated import deprecated


class _FFList:
    """
    A class represent a Fuzzy List of an element

    :Attributes:

        item: int
            the item name
        sumIUtil: float
            the sum of utilities of a fuzzy item in database
        sumRUtil: float
            the sum of resting values of a fuzzy item in database
        elements: list
            a list of elements contain tid,Utility and resting values of element in each transaction

    :Methods:

        addElement(element)
            Method to add an element to this fuzzy list and update the sums at the same time.
        printElement(e)
            Method to print elements            

    """

    def __init__(self, itemName):
        self.item = itemName
        self.sumIUtil = 0.0
        self.sumRUtil = 0.0
        self.elements = []

    def addElement(self, element):
        """
        A Method that add a new element to FFList

        :param element: an element to be added to FFList
        :type element: Element
        """
        self.sumIUtil += element.iUtils
        self.sumRUtil += element.rUtils
        self.elements.append(element)

    def printElement(self):
        """
        A Method to Print elements in the FFList
        """
        for ele in self.elements:
            print(ele.tid, ele.iUtils, ele.rUtils)


class _Element:
    """
    A class represents an Element of a fuzzy list

    :Attributes:

        tid : int
            keep tact of transaction id
        iUtils : float
            the utility of a fuzzy item in the transaction
        rUtils : float
            the neighbourhood resting value of a fuzzy item in the transaction
    """

    def __init__(self, tid, iUtil, rUtil):
        self.tid = tid
        self.iUtils = iUtil
        self.rUtils = rUtil


class _Regions:
    """
    A class calculate the regions

    :Attributes:

        low : int
            low region value
        middle : int
            middle region value
        high : int
            high region values
    """

    def __init__(self, quantity, regionsNumber):
        self.low = 0
        self.middle = 0
        self.high = 0
        if regionsNumber == 3:  # if we have 3 regions
            if 0 < quantity <= 1:
                self.low = 1
                self.high = 0
                self.middle = 0
            elif 1 < quantity <= 6:
                self.low = float((6 - quantity) / 5)
                self.middle = float((quantity - 1) / 5)
                self.high = 0
            elif 6 < quantity <= 11:
                self.low = 0
                self.middle = float((11 - quantity) / 5)
                self.high = float((quantity - 6) / 5)
            else:
                self.low = 0
                self.middle = 0
                self.high = 1


class _Pair:
    """
    A class to store item and it's quantity together
    """

    def __init__(self):
        self.item = 0
        self.quantity = 0


class FFSPMiner(_ab._fuzzySpatialFrequentPatterns):
    """
        About this algorithm
        ====================

        :Description:   Fuzzy Frequent Spatial Pattern-Miner is desired to find all Spatially frequent fuzzy patterns
                        which is on-trivial and challenging problem to its huge search space.we are using efficient pruning
                        techniques to reduce the search space.

        Reference:   Reference: P. Veena, B. S. Chithra, R. U. Kiran, S. Agarwal and K. Zettsu, "Discovering Fuzzy Frequent
                     Spatial Patterns in Large Quantitative Spatiotemporal databases," 2021 IEEE International Conference on Fuzzy Systems
                     (FUZZ-IEEE), 2021, pp. 1-8, doi: 10.1109/FUZZ45933.2021.9494594.

        :param  iFile: str :
                       Name of the Input file to mine complete set of frequent patterns
        :param  oFile: str :
                       Name of the output file to store complete set of frequent patterns
        :param  minSup: int or float or str :
                       The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count. Otherwise, it will be treated as float.
        :param maxPer: float :
                       The user can specify maxPer in count or proportion of database size. If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
        :param nFile: str :
                       Name of the input file to mine complete set of frequent patterns
        :param  sep: str :
                       This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


        :Attributes:

            iFile : file
                Name of the input file to mine complete set of fuzzy spatial frequent patterns
            oFile : file
                Name of the oFile file to store complete set of fuzzy spatial frequent patterns
            minSup : float
                The user given minimum support
            neighbors : map
                keep track of neighbours of elements
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            startTime : float
                To record the start time of the mining process
            endTime : float
                To record the completion time of the mining process
            itemsCnt : int
                To record the number of fuzzy spatial itemSets generated
            mapItemsLowSum : map
                To keep track of low region values of items
            mapItemsMidSum : map
                To keep track of middle region values of items
            mapItemsHighSum : map
                To keep track of high region values of items
            mapItemSum : map
                To keep track of sum of Fuzzy Values of items
            mapItemRegions : map
                To Keep track of fuzzy regions of item
            joinsCnt : int
                To keep track of the number of FFI-list that was constructed
            BufferSize : int
                represent the size of Buffer
            itemSetBuffer : list
                to keep track of items in buffer

        :Methods:

            mine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            save(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            convert(value)
                To convert the given user specified value
            FSFIMining( prefix, prefixLen, fsFim, minSup)
                Method generate FFI from prefix
            construct(px, py)
                A function to construct Fuzzy itemSet from 2 fuzzy itemSets
            Intersection(neighbourX,neighbourY)
                Return common neighbours of 2 itemSet Neighbours
            findElementWithTID(uList, tid)
                To find element with same tid as given
            WriteOut(prefix, prefixLen, item, sumIUtil,period)
                To Store the patten

        Execution methods
        =================


        **Terminal command**
        .. code-block:: console

          Format:

          (.venv) $ python3 FFSPMiner_old.py <inputFile> <outputFile> <neighbours> <minSup> <sep>

          Example Usage:

          (.venv) $ python3  FFSPMiner_old.py sampleTDB.txt output.txt sampleN.txt 3

          (.venv) $ python3  FFSPMiner_old.py sampleTDB.txt output.txt sampleN.txt 0.3

          (.venv) $ python3  FFSPMiner_old.py sampleTDB.txt output.txt sampleN.txt 3

        .. note:: minSup will be considered in percentage of database transactions


        **Sample run of importing the code:**
        ----------------------------------------

                from PAMI.fuzzyGeoreferencedFrequentPattern import FFSPMiner as alg

                obj = alg.FFSPMiner("input.txt", "neighbours.txt", 2)

                obj.mine()

                fuzzySpatialFrequentPatterns = obj.getPatterns()

                print("Total number of fuzzy frequent spatial patterns:", len(fuzzySpatialFrequentPatterns))

                obj.save("outputFile")

                memUSS = obj.getMemoryUSS()

                print("Total Memory in USS:", memUSS)

                memRSS = obj.getMemoryRSS()

                print("Total Memory in RSS", memRSS)

                run = obj.getRuntime()

                print("Total ExecutionTime in seconds:", run)

        Credits
        =======
                The complete program was written by B.Sai Chitra under the supervision of Professor Rage Uday Kiran.
    """
    _minSup = str()
    _iFile = " "
    _nFile = " "
    _sep = "\t"

    def __init__(self, iFile, nFile, minSup, sep="\t"):
        super().__init__(iFile, nFile, minSup, sep)
        self._mapItemNeighbours = {}
        self._startTime = 0
        self._endTime = 0
        self._itemsCnt = 0
        self._mapItemsLowSum = {}
        self._mapItemsMidSum = {}
        self._mapItemsHighSum = {}
        self._mapItemSum = {}
        self._mapItemRegions = {}
        self._joinsCnt = 0
        self._BufferSize = 200
        self._itemSetBuffer = []
        self._finalPatterns = {}
        self._dbLen = 0

    def _compareItems(self, o1, o2):
        """
        A Function that sort all FFI-list in ascending order of Support
        """
        compare = self._mapItemSum[o1.item] - self._mapItemSum[o2.item]
        if compare == 0:
            return int(o1.item) - int(o2.item)
        else:
            return compare

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value
        :type value: int or float or str
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._dbLen * value)
        if type(value) is str:
            if '.' in value:
                value = (self._dbLen * value)
            else:
                value = int(value)
        return value

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._transactions, self._fuzzyValues = [], []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.transactions = self._iFile['Transactions'].tolist()
            if 'fuzzyValues' in i:
                self.fuzzyValues = self._iFile['Utilities'].tolist()

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line = line.decode("utf-8")
                    line = line.split("\n")[0]
                    parts = line.split(":")
                    parts[0] = parts[0].strip()
                    parts[2] = parts[2].strip()
                    items = parts[0].split(self._sep)
                    quantities = parts[2].split(self._sep)
                    self.transactions.append([x for x in items])
                    self.fuzzyValues.append([x for x in quantities])
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.split("\n")[0]
                            parts = line.split(":")
                            parts[0] = parts[0].strip()
                            parts[2] = parts[2].strip()
                            items = parts[0].split(self._sep)
                            quantities = parts[2].split(self._sep)
                            self._transactions.append([x for x in items])
                            self._fuzzyValues.append([x for x in quantities])
                except IOError:
                    print("File Not Found")
                    quit()

    def _mapNeighbours(self):
        self._mapItemNeighbours = {}
        if isinstance(self._nFile, _ab._pd.DataFrame):
            data, items = [], []
            if self._nFile.empty:
                print("its empty..")
            i = self._nFile.columns.values.tolist()
            if 'items' in i:
                items = self._nFile['items'].tolist()
            if 'Neighbours' in i:
                data = self._nFile['Neighbours'].tolist()
            for k in range(len(items)):
                self._mapItemNeighbours[items[k]] = data[k]

        if isinstance(self._nFile, str):
            if _ab._validators.url(self._nFile):
                data = _ab._urlopen(self._nFile)
                for line in data:
                    line = line.decode("utf-8")
                    line = line.split("\n")[0]
                    parts = [i.rstrip() for i in line.split(self._sep)]
                    parts = [x for x in parts]
                    item = parts[0]
                    neigh1 = []
                    for i in range(1, len(parts)):
                        neigh1.append(parts[i])
                    self._mapItemNeighbours[item] = neigh1
            else:
                try:
                    with open(self._nFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.split("\n")[0]
                            parts = [i.rstrip() for i in line.split(self._sep)]
                            parts = [x for x in parts]
                            item = parts[0]
                            neigh1 = []
                            for i in range(1, len(parts)):
                                neigh1.append(parts[i])
                            self._mapItemNeighbours[item] = neigh1
                except IOError:
                    print("File Not Found")
                    quit()

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Frequent pattern mining process will start from here
        """
        self.mine()

    def mine(self):
        """
        Frequent pattern mining process will start from here
        """
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._finalPatterns = {}
        self._mapNeighbours()
        for line in range(len(self._transactions)):
            items = self._transactions[line]
            quantities = self._fuzzyValues[line]
            self._dbLen += 1
            for i in range(0, len(items)):
                regions = _Regions(int(quantities[i]), 3)
                item = items[i]
                if item in self._mapItemsLowSum.keys():
                    low = self._mapItemsLowSum[item]
                    low += regions.low
                    self._mapItemsLowSum[item] = low
                else:
                    self._mapItemsLowSum[item] = regions.low
                if item in self._mapItemsMidSum.keys():
                    mid = self._mapItemsMidSum[item]
                    mid += regions.middle
                    self._mapItemsMidSum[item] = mid
                else:
                    self._mapItemsMidSum[item] = regions.middle
                if item in self._mapItemsHighSum.keys():
                    high = self._mapItemsHighSum[item]
                    high += regions.high
                    self._mapItemsHighSum[item] = high
                else:
                    self._mapItemsHighSum[item] = regions.high
        listOfFFList = []
        mapItemsToFFLIST = {}
        self._minSup = self._convert(self._minSup)
        for item1 in self._mapItemsLowSum.keys():
            item = item1
            low = self._mapItemsLowSum[item]
            mid = self._mapItemsMidSum[item]
            high = self._mapItemsHighSum[item]
            if low >= mid and low >= high:
                self._mapItemSum[item] = low
                self._mapItemRegions[item] = "L"
            elif mid >= low and mid >= high:
                self._mapItemSum[item] = mid
                self._mapItemRegions[item] = "M"
            elif high >= low and high >= mid:
                self._mapItemRegions[item] = "H"
                self._mapItemSum[item] = high
            if self._mapItemSum[item] >= self._minSup:
                fuList = _FFList(item)
                mapItemsToFFLIST[item] = fuList
                listOfFFList.append(fuList)
        listOfFFList.sort(key=_ab._functools.cmp_to_key(self._compareItems))
        tid = 0
        for line in range(len(self._transactions)):
            items = self._transactions[line]
            quantities = self._fuzzyValues[line]
            revisedTransaction = []
            for i in range(0, len(items)):
                pair = _Pair()
                pair.item = items[i]
                regions = _Regions(int(quantities[i]), 3)
                item = pair.item
                if self._mapItemSum[item] >= self._minSup:
                    if self._mapItemRegions[pair.item] == "L":
                        pair.quantity = regions.low
                    elif self._mapItemRegions[pair.item] == "M":
                        pair.quantity = regions.middle
                    elif self._mapItemRegions[pair.item] == "H":
                        pair.quantity = regions.high
                    if pair.quantity > 0:
                        revisedTransaction.append(pair)
            revisedTransaction.sort(key=_ab._functools.cmp_to_key(self._compareItems))
            for i in range(len(revisedTransaction) - 1, -1, -1):
                pair = revisedTransaction[i]
                remainUtil = 0
                for j in range(len(revisedTransaction) - 1, i, -1):
                    if self._mapItemNeighbours.get(pair.item) is None:
                        continue
                    if revisedTransaction[j].item in self._mapItemNeighbours[pair.item]:
                        remainUtil += revisedTransaction[j].quantity
                remainingUtility = remainUtil
                if mapItemsToFFLIST.get(pair.item) is not None:
                    FFListOfItem = mapItemsToFFLIST[pair.item]
                    element = _Element(tid, pair.quantity, remainingUtility)
                    FFListOfItem.addElement(element)
            tid += 1
        itemNeighbours = list(self._mapItemNeighbours.keys())
        self._FSFIMining(self._itemSetBuffer, 0, listOfFFList, self._minSup, itemNeighbours)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def _FSFIMining(self, prefix, prefixLen, FSFIM, minSup, itemNeighbours):
        """
        Generates FFSPMiner from prefix

        :param prefix: the prefix patterns of FFSPMiner
        :type prefix: len
        :param prefixLen: the length of prefix
        :type prefixLen: int
        :param FSFIM: the Fuzzy list of prefix itemSets
        :type FSFIM: list
        :param minSup: the minimum support of
        :type minSup: int
        :param itemNeighbours: the set of common neighbours of prefix
        :type itemNeighbours: list or set
        """
        for i in range(0, len(FSFIM)):
            X = FSFIM[i]
            if X.sumIUtil >= minSup:
                self._WriteOut(prefix, prefixLen, X.item, X.sumIUtil)
            newNeighbours = self._Intersection(self._mapItemNeighbours.get(X.item), itemNeighbours)
            if X.sumRUtil >= minSup:
                exULs = []
                for j in range(i + 1, len(FSFIM)):
                    Y = FSFIM[j]
                    if Y.item in newNeighbours:
                        exULs.append(self._construct(X, Y))
                        self._joinsCnt += 1
                self._itemSetBuffer.insert(prefixLen, X.item)
                self._FSFIMining(self._itemSetBuffer, prefixLen + 1, exULs, minSup, newNeighbours)

    def _Intersection(self, neighbourX, neighbourY):
        """
        A function to get common neighbours from 2 itemSets

        :param neighbourX: the set of neighbours of itemSet 1
        :type neighbourX: set or list
        :param neighbourY: the set of neighbours of itemSet 2
        :type neighbourY: set or list
        :return: set of common neighbours of 2 itemSets
        :rtype: set
        """
        result = []
        if neighbourX is None or neighbourY is None:
            return result
        for i in range(0, len(neighbourX)):
            if neighbourX[i] in neighbourY:
                result.append(neighbourX[i])
        return result

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self._endTime - self._startTime

    def _construct(self, px, py):
        """
        A function to construct a new Fuzzy itemSet from 2 fuzzy itemSets

        :param px: the itemSet px
        :type px: FFI-List
        :param py: itemSet py
        :type py: FFI-List
        :return: the itemSet of pxy(px and py)
        :rtype: FFI-List
        """
        pxyUL = _FFList(py.item)
        for ex in px.elements:
            ey = self._findElementWithTID(py, ex.tid)
            if ey is None:
                continue
            eXY = _Element(ex.tid, min([ex.iUtils, ey.iUtils], key=lambda x: float(x)), ey.rUtils)
            pxyUL.addElement(eXY)
        return pxyUL

    def _findElementWithTID(self, uList, tid):
        """
        To find element with same tid as given

        :param uList: fuzzyList
        :type uList: FFI-List
        :param tid: transaction id
        :type tid: int
        :return: element tid as given
        :rtype: element if exist or None
        """
        List = uList.elements
        first = 0
        last = len(List) - 1
        while first <= last:
            mid = (first + last) >> 1
            if List[mid].tid < tid:
                first = mid + 1
            elif List[mid].tid > tid:
                last = mid - 1
            else:
                return List[mid]
        return None

    def _WriteOut(self, prefix, prefixLen, item, sumIUtil):
        """
        To Store the patten

        :param prefix: prefix of itemSet
        :type prefix: list
        :param prefixLen: length of prefix
        :type prefixLen: int
        :param item: the last item
        :type item: int
        :param sumIUtil: sum of utility of itemSet
        :type sumIUtil: float
        """
        self._itemsCnt += 1
        res = ""
        for i in range(0, prefixLen):
            res += str(prefix[i]) + "." + str(self._mapItemRegions[prefix[i]]) + "\t"
        res += str(item) + "." + str(self._mapItemRegions.get(item))
        res1 = str(sumIUtil)
        self._finalPatterns[res] = res1

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self._finalPatterns.items():
            patternsAndSupport = x.strip() + " : " + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Spatial Fuzzy Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in seconds:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = FFSPMiner(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = FFSPMiner(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.mine()
        _ap.mine()
        print("Total number of Spatial Fuzzy Frequent  Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS",  _ap.getMemoryRSS())
        print("Total ExecutionTime in seconds:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



from .edge import Edge
from .vertex import Vertex

class Graph:
    """
    Represents a graph structure composed of vertices and edges.

    Attributes:
        EMPTY_VERTEX_LIST (list): An empty list used as a default value for vertex lists.
        EMPTY_INTEGER_ARRAY (list): An empty list used as a default value for integer arrays.

    """
    EMPTY_VERTEX_LIST = []
    EMPTY_INTEGER_ARRAY = []

    def __init__(self, id, vMap=None, dfsCode=None):
        """
        Initializes the Graph object with optional parameters.
        """
        self.vMap = {}
        self.id = id
        if vMap is not None:
            self.vMap = vMap
        elif dfsCode is not None:
            for ee in dfsCode.getEeList():
                v1, v2, v1Label, v2Label, eLabel = ee.getV1(), ee.getV2(), ee.getVLabel1(), ee.getVLabel2(), ee.getEdgeLabel()
                
                e = Edge(v1, v2, eLabel)
                if v1 not in self.vMap:
                    self.vMap[v1] = Vertex(v1, v1Label)
                if v2 not in self.vMap:
                    self.vMap[v2] = Vertex(v2, v2Label)

                self.vMap[v1].addEdge(e)
                self.vMap[v2].addEdge(e)

            self.id = -1

        self.vertices = []
        self.neighborCache = {}
        self.mapLabelToVertexIds = {}
        self.edgeCount = 0

        self.precalculateVertexList()
        self.precalculateVertexNeighbors()
        self.precalculateLabelsToVertices()

    def getId(self):
        """
        Retrieves the ID of the graph.
        """
        return self.id

    def removeInfrequentLabel(self, label):
        """
        Removes vertices with infrequent labels and their associated edges.
        """
        toRemove = [key for key, vertex in self.vMap.items() if vertex.getLabel() == label]
        for key in toRemove:
            del self.vMap[key]

        for vertex in self.vMap.values():
            edgesToRemove = [edge for edge in vertex.getEdgeList() 
                               if edge.v1 not in self.vMap or edge.v2 not in self.vMap]

            for edge in edgesToRemove:
                vertex.getEdgeList().remove(edge)

    def precalculateVertexNeighbors(self):
        """
        Precalculates and caches the neighbor vertices for each vertex.
        """
        self.neighborCache = {}
        self.edgeCount = 0

        for vertexId, vertex in self.vMap.items():
            neighbors = []

            for edge in vertex.getEdgeList():
                neighborVertex = self.vMap[edge.another(vertexId)]
                neighbors.append(neighborVertex)

            neighbors.sort(key=lambda x: x.id)

            self.neighborCache[vertexId] = neighbors
            self.edgeCount += len(neighbors)

        self.edgeCount //= 2    
    
    def precalculateVertexList(self):
        """
        Precalculates and caches the list of vertices.
        """
        self.vertices = []
        for _, vertex in self.vMap.items():
            self.vertices.append(vertex)

    def precalculateLabelsToVertices(self):
        """
        Precalculates and caches a mapping of labels to vertex IDs.
        """
        self.mapLabelToVertexIds = {}
        for vertex in self.vertices:
            label = vertex.getLabel()
            if label not in self.mapLabelToVertexIds:
                sameIds = [v.getId() for v in self.vertices if v.getLabel() == label]
                self.mapLabelToVertexIds[label] = sameIds

    def findAllWithLabel(self, targetLabel):
        """
        Finds all vertices with a specified label.
        """
        if targetLabel in self.mapLabelToVertexIds:
            return self.mapLabelToVertexIds[targetLabel]
        else:
            return []
        
    def getAllNeighbors(self, v):
        """
        Retrieves all neighbors of a vertex.
        """
        try:
            neighbors = self.neighborCache[v]
        except KeyError:
            neighbors = []
        return neighbors
    
    def getVLabel(self, v):
        """
        Retrieves the label of a vertex.
        """
        return self.vMap[v].getLabel()
    
    def getEdgeLabel(self, v1, v2):
        """
        Retrieves the label of an edge between two vertices.
        """
        for e in self.vMap.get(v1).getEdgeList():
            if e.v1 == v1 and e.v2 == v2:
                return e.getEdgeLabel()
            elif e.v1 == v2 and e.v2 == v1:
                return e.getEdgeLabel()
        return -1
    

    def getEdge(self, v1, v2):
        """
        Retrieves the edge between two vertices.
        """
        for e in self.vMap.get(v1).getEdgeList():
            if e.v1 == v1 and e.v2 == v2:
                return e
            elif e.v1 == v2 and e.v2 == v1:
                return e
        return None
    
    def getNonPrecalculatedAllVertices(self):
        """
        Retrieves all vertices that have not been precalculated.

        """
        return list(self.vMap.values())
    
    def isNeighboring(self, v1, v2):
        """

        Checks if two vertices are neighbors.

        """
        neighborsOfV1 = self.neighborCache.get(v1, [])
        low = 0
        high = len(neighborsOfV1) - 1

        while high >= low:
            middle = (low + high) // 2
            val = neighborsOfV1[middle].id
            if val == v2:
                return True
            if val < v2:
                low = middle + 1
            if val > v2:
                high = middle - 1
        return False

    def getAllVertices(self):
        """
        Retrieves all vertices in the graph.
        """
        return self.vertices
    
    def getEdgeCount(self):
        """
        Retrieves the total number of edges in the graph.
        """
        return self.edgeCount



from abc import ABC, abstractmethod
from .graph import UncertainGraph
from .vertex import Vertex
from .edge import Edge
from .dfsCode import DFSCode

import math
import random
from copy import deepcopy

import itertools


class _MUSE(ABC):

        @abstractmethod
        def mine(self, minsup, epsilon, delta):
            """
            Run the muse algorithm.
            """
            pass

        @abstractmethod
        def readGraph(self, path):
            """
            Read graphs from a file.
            :param path: Path to the file containing the graphs.
            """
            pass

        @abstractmethod
        def save(self, oFile):
            """
            Write the result of the muse algorithm to a file.
            :param outputPath: Path to the output file.
            """
            pass




#  Copyright (C)  2024 Rage Uday Kiran , P Krishna Reddy
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.



from abc import ABC as ABC, abstractmethod as _abstractmethod
import time as _time
import validators as _validators
from urllib.request import urlopen as _urlopen
import csv as _csv
import pandas as _pd
from collections import defaultdict as _defaultdict
from itertools import combinations as _c
import os as _os
import os.path as _ospath
import psutil as _psutil
from array import *
import functools as _functools
import sys as _sys


class GTCP(ABC):
    def __init__(self,iFile,minsup,minGTC,minGTPC,maxOR=0.2):
        """
            iFile : input file
            minsup : Minimum support 
            minGTC : Minimum Graph transaction coverage
            minGTPC : Minimum graph pattern coverage 
            maxOR : Maximum overlap ratio
        """

    @_abstractmethod
    def mine(self):
        """
            Mine the coverage patterns
        """
    
    @_abstractmethod
    def getPatterns(self):
        """
            Get all the coverage patterns
        """

    @_abstractmethod    
    def Coverage(self,g):
        """
            Get coverage of a graph
            param
                g : Graph id
        """

    @_abstractmethod
    def patternCoverage(self,pattern):
        """
            Get patternCoverage of a pattern
            param
                pattern: pattern for which pattern coverage needs to be computed
        """

    @_abstractmethod
    def OverlapRatio(self,pattern):
        """
            Get Overlap ratio of a pattern
            param
                pattern: pattern for which overlap ratio needs to be computed
        """


    @_abstractmethod
    def GetFIDBasedFlatTransactions(self):
        """
            Convert into FID based transactions
        """

    @_abstractmethod
    def getallFreq1(self):
        """
            Get all the Patterns of size 1

        """

    @_abstractmethod
    def join(self,l1,l2):
        """
            Join two patterns
            Param 
                l1: Pattern 1
                l2: Pattern 2
        """

    @_abstractmethod
    def Cmine(self):
       """
        Cmine Algorithm for mining coverage patterns

       """









# PPPGrowth is fundamental approach to mine the partial periodic patterns in temporal database.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#         from PAMI.periodicFrequentPattern.basic import PPPGrowth as alg
#
#         obj = alg.PPPGrowth(iFile, minPS, period)
#
#         obj.mine()
#
#         partialPeriodicPatterns = obj.getPatterns()
#
#         print("Total number of partial periodic Patterns:", len(partialPeriodicPatterns))
#
#         obj.save(oFile)
#
#         Df = obj.getPatternInDf()
#
#         memUSS = obj.getMemoryUSS()
#
#         print("Total Memory in USS:", memUSS)
#
#         memRSS = obj.getMemoryRSS()
#
#         print("Total Memory in RSS", memRSS)
#
#         run = obj.getRuntime()
#
#         print("Total ExecutionTime in seconds:", run)
#



__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""


from PAMI.partialPeriodicPattern.basic import abstract as _abstract
from typing import List, Dict, Tuple, Set, Union, Any, Iterable, Generator
import validators as _validators
from urllib.request import urlopen as _urlopen
import sys as _sys
import pandas as pd
import numpy as np
from deprecated import deprecated

_minPS = float()
_period = float()
_lno = int()

class _Node(object):
    """
    A class used to represent the node of frequentPatternTree

    :**Attributes**:    - **item** (*int or None*) -- *Storing item of a node.*
                        - **timeStamps** (*list*) -- *To maintain the timestamps of a database at the end of the branch.*
                        - **parent** (*list*) -- *To maintain the parent of every node.*
                        - **children** (*list*) -- *To maintain the children of a node.*

    :**Methods**:    -**addChild(itemName)** -- *Storing the children to their respective parent nodes.*
    """

    def __init__(self, item, locations, parent=None):
        self.item = item
        self.locations = locations
        self.parent = parent
        self.children = {}

    def addChild(self, item, locations):
        """
        This method takes an item and locations as input, adds a new child node
        if the item does not already exist among the current node's children, or
        updates the locations of the existing child node if the item is already present.

        :param item: Represents the distinct item to be added as a child node.
        :type item: Any
        :param locations: Represents the locations associated with the item.
        :type locations: list
        :return: The child node associated with the item.
        :rtype: _Node
        """
        if item not in self.children:
            self.children[item] = _Node(item, locations, self)
        else:
            self.children[item].locations = locations + self.children[item].locations
            
        return self.children[item]

    def traverse(self):
        """
        This method constructs a transaction by traversing from the current node to the root node, collecting items along the way.

        :return: A tuple containing the transaction and the locations associated with the current node.
        :rtype: tuple(list, Any)
        """
        transaction = []
        locs = self.locations
        node = self.parent
        while node.parent is not None:
            transaction.append(node.item)
            node = node.parent
        return transaction[::-1], locs


class PPPGrowth(_abstract._partialPeriodicPatterns):
    """
    :Description:   3pgrowth is fundamental approach to mine the partial periodic patterns in temporal database.

    :Reference:   Discovering Partial Periodic Itemsets in Temporal Databases,SSDBM '17: Proceedings of the 29th International Conference on Scientific and Statistical Database ManagementJune 2017
                  Article No.: 30 Pages 1–6https://doi.org/10.1145/3085504.3085535

    :param  iFile: str :
                   Name of the Input file to mine complete set of frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  minPS: float:
                   Minimum partial periodic pattern...
    :param  period: float:
                   Minimum partial periodic...

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        iFile : file
            Name of the Input file or path of the input file
        oFile : file
            Name of the output file or path of the output file
        minPS: float or int or str
            The user can specify minPS either in count or proportion of database size.
            If the program detects the data type of minPS is integer, then it treats minPS is expressed in count.
            Otherwise, it will be treated as float.
            Example: minPS=10 will be treated as integer, while minPS=10.0 will be treated as float
        period: float or int or str
            The user can specify period either in count or proportion of database size.
            If the program detects the data type of period is integer, then it treats period is expressed in count.
            Otherwise, it will be treated as float.
            Example: period=10 will be treated as integer, while period=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            it represents the total no of transactions
        tree : class
            it represents the Tree class
        finalPatterns : dict
            it represents to store the patterns

    :Methods:

        mine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets()
            Scans the dataset or dataframes and stores in list format
        partialPeriodicOneItem()
            Extracts the one-frequent patterns from transactions
        updateTransactions()
            updates the transactions by removing the aperiodic items and sort the transactions with items
            by decreasing support
        buildTree()
            constrcuts the main tree by setting the root node as null
        mine()
            main program to mine the partial periodic patterns

    **Executing the code on terminal:**
    --------------------------------------
      .. code-block:: console


       Format:

       (.venv) $python3 PPPGrowth.py <inputFile> <outputFile> <minPS> <period>
    
       Examples:

       (.venv) $ python3 PPPGrowth.py sampleDB.txt patterns.txt 10.0 2.0


    **Sample run of the importing code:**
    -----------------------------------------
    .. code-block:: python

            from PAMI.periodicFrequentPattern.basic import PPPGrowth as alg

            obj = alg.PPPGrowth(iFile, minPS, period)

            obj.mine()

            partialPeriodicPatterns = obj.getPatterns()

            print("Total number of partial periodic Patterns:", len(partialPeriodicPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDf()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

                run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)


    **Credits:**
    -----------------
    The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

    """
    _minPS = float()
    _period = float()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _rank = {}
    _rankdup = {}
    _lno = 0

    def _creatingItemSets(self) -> None:
        """
        Storing the complete transactions of the database/input file in a database variable
        :return: None
        """
        self._Database = []
        if isinstance(self._iFile, pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                if data[i]:
                    tr = [str(ts[i])] + [x for x in data[i].split(self._sep)]
                    self._Database.append(tr)
                else:
                    self._Database.append([str(ts[i])])

        if isinstance(self._iFile, str):
            if _validators.url(self._iFile):
                data = _urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()


    def _convert(self, value: Union[int, float, str]) -> Union[int, float]:
        """
        To convert the given user specified value

        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._maxTS * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._maxTS * value)
            else:
                value = int(value)
        return value


    @deprecated("It is recommended to use mine() instead of mine() for mining process")
    def startMine(self) -> None:
        """
        Main method where the patterns are mined by constructing tree.
        :return: None
        """

        self.mine()

    def _getPerSup(self, arr):
        arr = list(arr)
        arr = np.sort(arr)
        arr = np.diff(arr)
        locs = len(np.where(arr <= self._period)[0])

        return locs
    

    def _construct(self, items, data):

        """
        This method filters the items based on the minimum support (minSup) and
        maximum period (maxPer). It then constructs a tree structure from the
        filtered items and data.

        :param items: A dictionary where keys are items and values are lists of timestamps.
        :type items: dict
        :param data: The dataset used to construct the tree, where each entry is a list with
                     an index followed by items.
        :type data: list of lists
        :param minSup: The minimum support threshold.
        :type minSup: int
        :param maxPer: The maximum period threshold.
        :type maxPer: int or float
        :param maxTS: The maximum timestamp.
        :type maxTS: int or float
        :param patterns: A dictionary to store the patterns discovered during the construction.
        :type patterns: dict
        :return: A tuple containing the root node of the constructed tree and a dictionary
                 of item nodes.
        :rtype: tuple(_Node, dict)
        """


        items = {k: v for k, v in items.items() if self._getPerSup(v) >= self._minPS}

        #tested ok
        for item, ts in items.items():
            self._finalPatterns[tuple([item])] = self._getPerSup(ts)

        root = _Node([], None, None)
        itemNodes = {}
        for line in data:
            currNode = root
            index = int(line[0])
            line = line[1:]
            line = sorted([item for item in line if item in items], key = lambda x: len(items[x]), reverse = True)
            for item in line:
                currNode = currNode.addChild(item, [index])   # heavy
                if item in itemNodes:
                    itemNodes[item].add(currNode)
                else:
                    itemNodes[item] = set([currNode])

        return root, itemNodes

    
    def _recursive(self, root, itemNode):
        """
        This method recursively constructs a pattern tree from the given root node,
        filtering items based on the minimum support (minSup) and maximum period (maxPer).
        It updates the patterns dictionary with the discovered patterns.

        :param root: The current root node of the pattern tree.
        :type root: _Node
        :param itemNode: A dictionary where keys are items and values are sets of nodes
                         associated with those items.
        :type itemNode: dict
        :param minSup: The minimum support threshold.
        :type minSup: int
        :param maxPer: The maximum period threshold.
        :type maxPer: int or float
        :param patterns: A dictionary to store the patterns discovered during the recursion.
        :type patterns: dict
        :param maxTS: The maximum timestamp.
        :type maxTS: int or float
        """

        for item in itemNode:
            newRoot = _Node(root.item + [item], None, None)

            itemLocs = {}
            transactions = {}
            for node in itemNode[item]:
                transaction, locs = node.traverse()
                if len(transaction) < 1:
                    continue
                # transactions.append((transaction, locs))
                if tuple(transaction) in transactions:
                    transactions[tuple(transaction)].extend(locs)
                else:
                    transactions[tuple(transaction)] = locs

                for item in transaction:
                    if item in itemLocs:
                        itemLocs[item] += locs
                    else:
                        itemLocs[item] = list(locs)

            # Precompute getMaxPer results for itemLocs
            # maxPerResults = {item: self._getMaxPer(itemLocs[item], maxTS) for item in itemLocs if len(itemLocs[item]) >= minSup}
            maxPerResults = {item: self._getPerSup(itemLocs[item]) for item in itemLocs}

            # Filter itemLocs based on minSup and maxPer
            itemLocs = {k: len(v) for k, v in itemLocs.items() if maxPerResults[k] >= self._minPS}

            # Iterate over filtered itemLocs
            for item in itemLocs:
                self._finalPatterns[tuple(newRoot.item + [item])] = maxPerResults[item]
            
            if not itemLocs:
                continue

            newItemNodes = {}

            for transaction, locs in transactions.items():
                transaction = sorted([item for item in transaction if item in itemLocs], key = lambda x: itemLocs[x], reverse = True)
                if len(transaction) < 1:
                    continue
                currNode = newRoot
                for item in transaction:
                    currNode = currNode.addChild(item, locs)
                    if item in newItemNodes:
                        newItemNodes[item].add(currNode)
                    else:
                        newItemNodes[item] = set([currNode])

            self._recursive(newRoot, newItemNodes)


    def mine(self) -> None:
        """
        Main method where the patterns are mined by constructing tree.
        :return: None

        """
        global _minPS, _period, _lno
        self._startTime = _abstract._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minPS is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        

        self._maxTS = 0
        items = {}
        for line in self._Database:
            index = int(line[0])
            self._maxTS = max(self._maxTS, index)
            for item in line[1:]:
                if item not in items:
                    items[item] = []
                items[item].append(index)

        self._minPS = self._convert(self._minPS)
        self._period = self._convert(self._period)

        root, itemNodes = self._construct(items, self._Database)
        self._recursive(root, itemNodes)

        newPattern = {}
        for k, v in self._finalPatterns.items():
            newPattern["\t".join([str(x) for x in k])] = v

        self._finalPatterns = newPattern


        self._endTime = _abstract._time.time()
        process = _abstract._psutil.Process(_abstract._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Partial Periodic Patterns were generated successfully using 3PGrowth algorithm ")


    def getMemoryUSS(self) -> float:
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self) -> float:
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self) -> _abstract._pd.DataFrame:
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        # dataFrame = {}
        # data = []
        # for a, b in self._finalPatterns.items():
        #     data.append([a.replace('\t', ' '), b])
        #     dataFrame = _abstract._pd.DataFrame(data, columns=['Patterns', 'periodicSupport'])
        # return dataFrame
        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
        dataFrame = pd.DataFrame(data, columns=['Patterns', 'Periodicity'])
        return dataFrame

    def save(self, outFile: str) -> None:
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: csv file
        :return: None
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self) -> Dict[str, int]:
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self) -> None:
        """
        This function is used to print the results
        :return: None
        """
        print("Total number of Partial Periodic Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_sys.argv) == 5 or len(_sys.argv) == 6:
        if len(_sys.argv) == 6:
            _ap = PPPGrowth(_sys.argv[1], _sys.argv[3], _sys.argv[4], _sys.argv[5])
        if len(_sys.argv) == 5:
            _ap = PPPGrowth(_sys.argv[1], _sys.argv[3], _sys.argv[4])
        _ap.mine()
        print("Total number of Partial Periodic Patterns:", len(_ap.getPatterns()))
        _ap.save(_sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:",  _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")
        for i in [100, 200, 300, 400, 500]:
            _ap = PPPGrowth('/Users/tarunsreepada/Downloads/Temporal_T10I4D100K.csv', i, 5000, '\t')
            _ap.mine()
            print("Total number of Maximal Partial Periodic Patterns:", len(_ap.getPatterns()))
            _ap.save('/Users/tarunsreepada/Downloads/output.txt')
            print(_ap.getPatternsAsDataFrame())
            print("Total Memory in USS:", _ap.getMemoryUSS())
            print("Total Memory in RSS", _ap.getMemoryRSS())
            print("Total ExecutionTime in ms:", _ap.getRuntime())


# MaxFP-Growth is one of the fundamental algorithm to discover maximal frequent patterns in a transactional database.
#
# **Importing this algorithm into a python program**
# ---------------------------------------------------------
#
#             from PAMI.frequentPattern.maximal import MaxFPGrowth as alg
#
#             obj = alg.MaxFPGrowth("../basic/sampleTDB.txt", "2")
#
#             obj.mine()
#
#             frequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#             obj.save("patterns")
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from PAMI.frequentPattern.maximal import abstract as _ab
from deprecated import deprecated


_minSup = str()
global maximalTree


class _Node(object):
    """
    A class used to represent the node of frequentPatternTree

    :Attributes:
        item : int
            storing item of a node
        counter : list
            To maintain the support of the node
        parent : node
            To maintain the parent of every node
        children : list
            To maintain the children of node

    :Methods:
        addChild(itemName)
            storing the children to their respective parent nodes
    """

    def __init__(self, item, children):
        """
        Initializing the Node class
        :param item: Storing the item of a node
        :type item: int or None
        :param children: To maintain the children of a node
        :type children: dict
        """
        self.item = item
        self.children = children
        self.counter = int()
        self.parent = None

    def addChild(self, node):
        """
        Adding a child to the created nodes
        :param node: node object
        :type node: Node
        """
        self.children[node.item] = node
        node.parent = self


class _Tree(object):
    """
    A class used to represent the frequentPatternGrowth tree structure

    :Attributes:
        root : Node
               Represents the root node of the tree
        summaries : dictionary
                storing the nodes with same item name
        info : dictionary
                stores the support of items


    :Methods:

        addTransaction(transaction)
            creating transaction as a branch in frequentPatternTree
        addConditionalTransaction(prefixPaths, supportOfItems)
            construct the conditional tree for prefix paths
        condPatterns(Node)
            generates the conditional patterns from tree for specific node
        conditionalTransaction(prefixPaths,Support)
            takes the prefixPath of a node and support at child of the path and extract the frequent items from
            prefixPaths and generates prefixPaths with items which are frequent
        remove(Node)
            removes the node from tree once after generating all the patterns respective to the node
        generatePatterns(Node)
            starts from the root node of the tree and mines the frequent patterns
    """

    def __init__(self):
        self.root = _Node(None, {})
        self.summaries = {}
        self.info = {}
        #self.maximalTree = _MPTree()

    def addTransaction(self, transaction):
        """

        Adding transactions into tree

        :param transaction: represents the transaction in a database

        :type transaction: list

        :return: tree

        """
        currentNode = self.root
        for i in range(len(transaction)):
            if transaction[i] not in currentNode.children:
                newNode = _Node(transaction[i], {})
                newNode.counter = 1
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]
                currentNode.counter += 1

    def addConditionalTransaction(self, transaction, count):
        """

        Loading the database into a tree

        :param transaction: conditional transaction of a node

        :type transaction: list

        :param count: the support of conditional transaction

        :type count: int

        :return: conditional tree

        """
        currentNode = self.root
        for i in range(len(transaction)):
            if transaction[i] not in currentNode.children:
                newNode = _Node(transaction[i], {})
                newNode.counter = count
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]
                currentNode.counter += count

    def getConditionalPatterns(self, alpha):
        """
        Generates all the conditional patterns of respective node
        :param alpha: it represents the Node in tree
        :type alpha: int
        :return: conditional patterns of a node
        """
        finalPatterns = []
        finalSets = []
        for i in self.summaries[alpha]:
            set1 = i.counter
            set2 = []
            while i.parent.item is not None:
                set2.append(i.parent.item)
                i = i.parent
            if len(set2) > 0:
                set2.reverse()
                finalPatterns.append(set2)
                finalSets.append(set1)
        finalPatterns, finalSets, info = self.conditionalTransactions(finalPatterns, finalSets)
        return finalPatterns, finalSets, info

    def conditionalTransactions(self, condPatterns, condFreq):
        """
        sorting and removing the items from conditional transactions which don't satisfy minSup
        :param condPatterns: conditional patterns if a node
        :type condPatterns: list
        :param condFreq: frequency at leaf node of conditional transaction
        :type condFreq: int
        :return: conditional patterns and their frequency respectively
        """
        global _minSup
        pat = []
        tids = []
        data1 = {}
        for i in range(len(condPatterns)):
            for j in condPatterns[i]:
                if j not in data1:
                    data1[j] = condFreq[i]
                else:
                    data1[j] += condFreq[i]
        updatedDict = {}
        updatedDict = {k: v for k, v in data1.items() if v >= _minSup}
        count = 0
        for p in condPatterns:
            p1 = [v for v in p if v in updatedDict]
            trans = sorted(p1, key=lambda x: (updatedDict.get(x), -x), reverse=True)
            if len(trans) > 0:
                pat.append(trans)
                tids.append(condFreq[count])
            count += 1
        return pat, tids, updatedDict

    def removeNode(self, nodeValue):
        """
        To remove the node from the original tree
        :param nodeValue: leaf node of tree
        :type nodeValue: int
        :return: tree after deleting node
        """
        for i in self.summaries[nodeValue]:
            del i.parent.children[nodeValue]
            i = None

    def generatePatterns(self, prefix, patterns, maximalTree):
        """
        Generates the patterns
        :param prefix: forms the combination of items
        :type prefix: str
        :param patterns: the patterns we want to generate for this node
        :type patterns: list
        :param maximalTree: maximal frequent patterns
        :type maximalTree: _MPTree()
        :return: the maximal frequent patterns
        :rtype: list
        """
        for i in sorted(self.summaries, key=lambda x: (self.info.get(x), -x)):
            pattern = prefix[:]
            pattern.append(i)
            condPatterns, tids, info = self.getConditionalPatterns(i)
            conditional_tree = _Tree()
            conditional_tree.info = info.copy()
            head = pattern[:]
            tail = []
            for la in info:
                tail.append(la)
            sub = head + tail
            if maximalTree.checkerSub(sub) == 1:
                for pat in range(len(condPatterns)):
                    conditional_tree.addConditionalTransaction(condPatterns[pat], tids[pat])
                if len(condPatterns) >= 1:
                    conditional_tree.generatePatterns(pattern, patterns, maximalTree)
                else:
                    maximalTree.addTransaction(pattern)
                    patterns[tuple(pattern)] = self.info[i]
            self.removeNode(i)


class _MNode(object):
    """
    A class used to represent the node in maximal tree

    :Attributes:
        item : int
            storing item of a node
        children : list
            To maintain the children of node

    :Methods:
        addChild(itemName)
            storing the children to their respective parent nodes
    """

    def __init__(self, item, children):
        self.item = item
        self.children = children

    def addChild(self, node):
        """
        To add the children details to a parent node
        :param node: children node
        :type node: _MNode
        :return: adding children details to parent node
        """
        self.children[node.item] = node
        node.parent = self


class _MPTree(object):
    """
    A class used to represent the frequentPatternGrowth tree structure

    :Attributes:

        root : Node
            Represents the root node of the tree
        summaries : dictionary
            storing the nodes with same item name


    :Methods:

        addTransaction(transaction)
            creating transaction as a branch in frequentPatternTree
        addConditionalTransaction(prefixPaths, supportOfItems)
            construct the conditional tree for prefix paths
        checkerSub(items):
            Given a set of items to the subset of them is present or not
    """

    def __init__(self):
        self.root = _MNode(None, {})
        self.summaries = {}

    def addTransaction(self, transaction):
        """
        To construct the maximal frequent pattern into maximal tree
        :param transaction: the maximal frequent patterns extracted till now
        :type transaction: list
        :return: the maximal tree
        """
        currentNode = self.root
        transaction.sort()
        for i in range(len(transaction)):
            if transaction[i] not in currentNode.children:
                newNode = _MNode(transaction[i], {})
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].insert(0, newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]

    def checkerSub(self, items):
        """
        To check the subset of pattern present in tree
        :param items: the sub frequent pattern
        :type items: list
        :return: checks if subset present in the tree
        """
        items.sort(reverse=True)
        item = items[0]
        if item not in self.summaries:
            return 1
        else:
            if len(items) == 1:
                return 0
        for t in self.summaries[item]:
            cur = t.parent
            i = 1
            while cur.item is not None:
                if items[i] == cur.item:
                    i += 1
                    if i == len(items):
                        return 0
                cur = cur.parent
        return 1


# Initialising the  variable for maximal tree
#maximalTree = _MPTree()


class MaxFPGrowth(_ab._frequentPatterns):
    """
    :Description: MaxFP-Growth is one of the fundamental algorithm to discover maximal frequent patterns in a transactional database.

    :Reference:   Grahne, G. and Zhu, J., "High Performance Mining of Maximal Frequent itemSets",
                  http://users.encs.concordia.ca/~grahne/papers/hpdm03.pdf

    :param  iFile: str :
                   Name of the Input file to mine complete set of frequent patterns
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  minSup: int or float or str :
                   The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
    :param  maxPer: float :
                   The user can specify maxPer in count or proportion of database size. If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.



    :Attributes:

        startTime : float
          To record the start time of the mining process

        endTime : float
          To record the completion time of the mining process

        finalPatterns : dict
          Storing the complete set of patterns in a dictionary variable

        memoryUSS : float
          To store the total amount of USS memory consumed by the program

        memoryRSS : float
          To store the total amount of RSS memory consumed by the program

        Database : list
          To store the transactions of a database in list

        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            it represents the total no of transactions
        tree : class
            it represents the Tree class
        itemSetCount : int
            it represents the total no of patterns
        finalPatterns : dict
            it represents to store the patterns


    **Methods to execute code on terminal**
    ---------------------------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 MaxFPGrowth.py <inputFile> <outputFile> <minSup>

      Example Usage:

      (.venv) $ python3 MaxFPGrowth.py sampleDB.txt patterns.txt 0.3

    .. note:: minSup will be considered in percentage of database transactions



    **Importing this algorithm into a python program**
    ---------------------------------------------------------

    .. code-block:: python

            from PAMI.frequentPattern.maximal import MaxFPGrowth as alg

            obj = alg.MaxFPGrowth("../basic/sampleTDB.txt", "2")

            obj.mine()

            frequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.savePatterns("patterns")

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    **Credits:**
    -------------------
                The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.

    """
    _startTime = float()
    _endTime = float()
    _minSup = str()
    _maxPer = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _rank = {}
    _rankdup = {}
    _lno = 0
    _maximalTree = str()

    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._Database = self._iFile['Transactions'].tolist()
                self._Database = [x.split(self._sep) for x in self._Database]
            else:
                print("The column name should be Transactions and each line should be separated by tab space or a seperator specified by the user")

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            #print(line)
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _frequentOneItem(self):
        """
        To extract the one-length frequent itemSets
        :return: 1-length frequent items
        """
        _mapSupport = {}
        k = 0
        for tr in self._Database:
            k += 1
            for i in range(0, len(tr)):
                if tr[i] not in _mapSupport:
                    _mapSupport[tr[i]] = 1
                else:
                    _mapSupport[tr[i]] += 1
        _mapSupport = {k: v for k, v in _mapSupport.items() if v >= self._minSup}
        #print(len(mapSupport), self.minSup)
        genList = [k for k, v in sorted(_mapSupport.items(), key=lambda x: x[1], reverse=True)]
        self._rank = dict([(index, item) for (item, index) in enumerate(genList)])
        return _mapSupport, genList

    def _updateTransactions(self, oneLength):
        """
        To sort the transactions in their support descending order and allocating ranks respectively
        :param oneLength: 1-length frequent items in dictionary
        :type oneLength: dict
        :return: returning the sorted list
        :Example: oneLength = {'a':7, 'b': 5, 'c':'4', 'd':3}
                    rank = {'a':0, 'b':1, 'c':2, 'd':3}
        """
        list1 = []
        for tr in self._Database:
            list2 = []
            for i in range(0, len(tr)):
                if tr[i] in oneLength:
                    list2.append(self._rank[tr[i]])
            if len(list2) >= 2:
                list2.sort()
                list1.append(list2)
        return list1


    def _buildTree(self, data, info):
        """
        creating the root node as null in fp-tree and adding all transactions into tree.
        :param data: updated transactions
        :type data: dict
        :param info: rank of items in transactions
        :type info: dict
        :return: fp-tree
        """
        rootNode = _Tree()
        rootNode.info = info.copy()
        for i in range(len(data)):
            rootNode.addTransaction(data[i])
        return rootNode


    def _convert(self, value):
        """
        To convert the type of user specified minSup value
        :param value: user specified minSup value
        :type value: int or float or str
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = ((len(self._Database)) * value)
            else:
                value = int(value)
        return value

    def _convertItems(self, itemSet):
        """
        To convert the item ranks into their original item names
        :param itemSet: itemSet or a pattern
        :type itemSet: list
        :return: original pattern
        """
        t1 = []
        for i in itemSet:
            t1.append(self._rankdup[i])
        return t1

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Mining process will start from this function
        """

        self.mine()

    def mine(self):
        """
        Mining process will start from this function
        """

        global _minSup
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minSup is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        _minSup = self._minSup
        generatedItems, pfList = self._frequentOneItem()
        updatedTransactions = self._updateTransactions(generatedItems)
        for x, y in self._rank.items():
            self._rankdup[y] = x
        info = {self._rank[k]: v for k, v in generatedItems.items()}
        patterns = {}
        self._finalPatterns = {}
        self._maximalTree = _MPTree()
        Tree = self._buildTree(updatedTransactions, info)
        Tree.generatePatterns([], patterns, self._maximalTree)
        for x, y in patterns.items():
            pattern = str()
            x = self._convertItems(x)
            for i in x:
                pattern = pattern + i + "\t"
            self._finalPatterns[pattern] = y
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Maximal Frequent patterns were generated successfully using MaxFp-Growth algorithm ")

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """

        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        # dataframe = {}
        # data = []
        # for a, b in self._finalPatterns.items():
        #     data.append([a.replace('\t', ' '), b])
        #     dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])

        dataframe = _ab._pd.DataFrame(list([[x.replace('\t', ' '), y] for x,y in self._finalPatterns.items()]), columns=['Patterns', 'Support'])

        return dataframe

    def save(self, outFile):
        """

        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csvfile
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns
    
    def printResults(self):
        """
        This functon is used to print the results
        """
        print('Total number of Maximal Frequent Patterns: ' + str(len(self.getPatterns())))
        print('Runtime: ' + str(self.getRuntime()))
        print('Memory (RSS): ' + str(self.getMemoryRSS()))
        print('Memory (USS): ' + str(self.getMemoryUSS()))


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = MaxFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = MaxFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.mine()
        _ap.mine()
        _ap.save(_ab._sys.argv[2])
        print("Total number of Maximal Frequent Patterns:", len(_ap.getPatterns()))
        print("Total Memory in USS:",  _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")






from PAMI.uncertainGraphMining.muse import abstract as _ab
class Muse(_ab._MUSE):
    def __init__(self, file_path):
        self.graphDatabase = self.readGraph(file_path)
        self.F = set()


    def readGraph(self, path):
        with open(path, 'r') as f:
            graphDatabase = []
            vMap = {}
            gId = None

            for line in f:
                line = line.strip()

                if line.startswith("t"):
                    if vMap:  # If vMap is not empty, it means a graph was read
                        graphDatabase.append(_ab.UncertainGraph(gId, vMap))
                        vMap = {}  # Reset for the next graph

                    gId = int(line.split(" ")[2])

                elif line.startswith("v"):
                    items = line.split(" ")
                    vId = int(items[1])
                    vLabel = int(items[2])
                    vMap[vId] = _ab.Vertex(vId, vLabel)

                elif line.startswith("e"):
                    items = line.split(" ")
                    v1 = int(items[1])
                    v2 = int(items[2])
                    eLabel = int(items[3])
                    eProb = float(items[4])
                    e = _ab.Edge(v1, v2, eLabel, eProb)
                    vMap[v1].addEdge(e)
                    vMap[v2].addEdge(e)

            if vMap:
                graphDatabase.append(_ab.UncertainGraph(gId, vMap))

        self.graphCount = len(graphDatabase)
        return graphDatabase

    def _generateImplicatedGraphs(self):
        implicatedGraphs = []
        for uncertainGraph in self.graphDatabase:
            vMap = uncertainGraph.getVertexMap()
            edges = [(e.v1, e.v2, e.edgeLabel, e.existenceProbability) for v in vMap.values() for e in v.getEdgeList()]
            edges = list(set(edges))  # Remove duplicates

            breakpoint()
            edgeCombinations = list(_ab.itertools.product([0, 1], repeat=len(edges)))

            for combination in edgeCombinations:
                implicatedGraph = {
                    "uncertain_graph_id": uncertainGraph.id,
                    "vertices": {vId: vertex.getLabel() for vId, vertex in vMap.items()},
                    "edges": [],
                    "implication_probability": 1
                }

                implicationProbability = 1
                for i, include_edge in enumerate(combination):
                    v1, v2, eLabel, eProb = edges[i]
                    if include_edge:
                        implicatedGraph["edges"].append((v1, v2, eLabel))
                        implicationProbability *= eProb
                    else:
                        implicationProbability *= (1 - eProb)

                implicatedGraph["implication_probability"] = implicationProbability
                implicatedGraphs.append(implicatedGraph)

        return implicatedGraphs

    def _generateImplicatedDatabase(self):
        allImplicatedGraphs = [self._generateImplicatedGraphs() for _ in self.graphDatabase]

        # Combine implicated graphs across all uncertain graphs using Cartesian product
        implicatedDatabase = {}
        counter = 0

        breakpoint()
        for combination in _ab.itertools.product(*allImplicatedGraphs):
            combinedGraph = {
                "vertices": {},
                "edges": [],
                "implication_probability": 1,
                "component_implicated_graphs": []
            }

            for graph in combination:
                combinedGraph["vertices"].update(graph["vertices"])
                combinedGraph["edges"].extend(graph["edges"])
                combinedGraph["implication_probability"] *= graph["implication_probability"]
                combinedGraph["component_implicated_graphs"].append(graph["uncertain_graph_id"])

            if all(combinedGraph["component_implicated_graphs"][i] < combinedGraph["component_implicated_graphs"][
                i + 1] for i in range(len(combinedGraph["component_implicated_graphs"]) - 1)):
                implicatedDatabase[f'X{counter}'] = combinedGraph
                counter += 1

        return implicatedDatabase

    def _sampleGraph(self, uncertainGraph):
        sampledGraph = _ab.deepcopy(uncertainGraph)
        for vertex in sampledGraph.getVertexMap().values():
            edgesToRemove = []
            for edge in vertex.getEdgeList():
                if _ab.random.random() > edge.existenceProbability:
                    edgesToRemove.append(edge)
            for edge in edgesToRemove:
                vertex.edges.remove(edge)
        return sampledGraph

    def _checkSubgraph(self, graph, subgraph):
        """ Check if subgraph is present in graph using DFS Code for subgraph isomorphism check """
        subgraphCode = _ab.DFSCode()
        subgraphCode.buildCode(subgraph)
        graphCode = _ab.DFSCode()
        graphCode.buildCode(graph)

        return subgraphCode.isSubcodeOf(graphCode)

    def _constructDnfFormula(self, embeddings):
        """ Construct the DNF formula from embeddings """
        dnfFormula = []
        for embedding in embeddings:
            prob = 1
            for edge in embedding.getEdgeList():
                prob *= edge.getExistenceProbability()
            dnfFormula.append(prob)
        return dnfFormula

    def estimateProbability(self, dnfFormula, epsilon, delta, n, minsup):
        """ Estimate the probability using FPRAS """
        epsilonPrime = epsilon * minsup / 2
        N = _ab.math.ceil((4 * n * _ab.math.log(2 / delta)) / (epsilonPrime ** 2))
        Z = sum(dnfFormula)
        X = Y = 0

        for _ in range(N):
            i = _ab.random.randint(0, n - 1)
            Ci = dnfFormula[i]
            Y += Ci
            if all(_ab.random.random() > dnfFormula[j] for j in range(n) if j != i):
                X += Ci

        pHat = (X * Z) / Y if Y > 0 else 0
        return max(0, pHat - epsilonPrime), min(1, pHat + epsilonPrime)

    def _approxOccProb(self, uncertainGraph, subgraph, epsilon, delta, minsup):
        embeddings = []
        for node in uncertainGraph.getVertexMap().values():
            if self._checkSubgraph(uncertainGraph.getVertexMap(), subgraph):
                embeddings.append(node)  # Collect embeddings as Vertex objects

        n = len(embeddings)
        if n == 0:
            return 0, 0

        dnfFormula = self._constructDnfFormula(embeddings)
        return self.estimateProbability(dnfFormula, epsilon, delta, n, minsup)

    def _computeProbabilityFromDnf(self, dnfFormula):
        """ Compute the probability of the DNF formula being true """
        return sum(dnfFormula) / len(dnfFormula) if dnfFormula else 0

    def _approxExpSup(self, subgraph, database, minsup, epsilon, delta):
        l = u = 0
        for Gi in database:
            Xi = [vertex for vertex in Gi.getVertexMap().values() if
                  self._checkSubgraph(Gi.getVertexMap(), subgraph)]
            if len(Xi) > 0 and 2 * len(Xi) - 5 / len(Xi) >= _ab.math.log(2 / delta) / (epsilon * minsup) ** 2:
                alpha, beta = self._approxOccProb(Gi, subgraph, epsilon, delta, minsup)
            else:
                dnfFormula = self._constructDnfFormula(Xi)
                alpha = beta = self._computeProbabilityFromDnf(dnfFormula)
            l += alpha
            u += beta
        n = len(database)
        return l / n, u / n

    def _generateInitialPatterns(self):
        patterns = set()
        for graph in self.graphDatabase:
            vertex_map = graph.getVertexMap()
            for vertex in vertex_map.values():
                for edge in vertex.getEdgeList():
                    pattern = {
                        vertex.getId(): vertex,
                        edge.v2: _ab.Vertex(edge.v2, vertex_map[edge.v2].getLabel())  # Access the label from vertex_map
                    }
                    patterns.add(frozenset(pattern.items()))
        return patterns

    def _generateSuperpatterns(self, pattern):
        superpatterns = set()
        for graph in self.graphDatabase:
            for vertex in graph.getVertexMap().values():
                for edge in vertex.getEdgeList():
                    if (vertex.id, vertex.getLabel()) in pattern:
                        new_pattern = dict(pattern)
                        if edge not in new_pattern[(vertex.id, vertex.getLabel())]:
                            new_pattern[(vertex.id, vertex.getLabel())].append(edge)
                            superpatterns.add(frozenset(new_pattern.items()))
        return superpatterns

    def mine(self, minsup, epsilon, delta):
        T = self._generateInitialPatterns()
        while T:
            S = T.pop()
            S_dict = dict(S)
            embeddings = []
            for Gi in self.graphDatabase:
                if self._checkSubgraph(Gi.getVertexMap(), S_dict):
                    embeddings.append(S_dict)
            l, u = self._approxExpSup(S_dict, self.graphDatabase, minsup, epsilon, delta)
            if l >= (1 - epsilon) * minsup and u >= minsup:
                self.F.add(S)
                superpatterns = self._generateSuperpatterns(S)
                for sp in superpatterns:
                    T.add(sp)

        return self.F

    def parseFrozenset(self, fset):
        vertices = {}
        edges = []

        edge_list = list(fset)
        for eid, vertex in edge_list:
            if vertex.id not in vertices:
                vertices[vertex.id] = vertex.vLabel
            if len(edge_list) > 1:
                for other_eid, other_vertex in edge_list:
                    if other_vertex.id != vertex.id:
                        edges.append((vertex.id, other_vertex.id))

        return vertices, edges
    def save(self, oFile):
        with open(oFile, 'w') as f:
            for tid, fset in enumerate(self.F):
                f.write(f"t # {tid}\n")
                vertices, edges = self.parseFrozenset(fset)
                for vid, label in vertices.items():
                    f.write(f"v {vid} {label}\n")

                for edge in edges:
                    f.write(f"e {edge[0]} {edge[1]}\n")



if __name__ == '__main__':
    muse = Muse('largerDataset')
    epsilon = 0.1
    delta = 0.05
    minsup = 0.3

    frequentPatterns = muse.mine(minsup, epsilon, delta)
    # print(f"Frequent subgraph patterns: {frequentPatterns}")
    for pattern in frequentPatterns:
        print(pattern)


import random as _rd
import psutil,os,time

class syntheticUtilityDatabase:
    """
    This class creates a synthetic utility database.

    Attributes:
        totalTransactions :int
                    Number of transactions.
        numOfItems : int
                    Number of items.
        maxUtilRange : int
                    Maximum utility range.
        avgTransactionLength : int
                    The length of average transaction.
        memoryUSS : float
                    To store the total amount of USS memory consumed by the program
        memoryRSS : float
                    To store the total amount of RSS memory consumed by the program
        startTime : float
                    To record the start time of the mining process
        endTime : float
                    To record the completion time of the mining process

    Methods:
        __init__(totalTransactions, numOfItems, maxUtilRange, avgTransactionLength)
            Constructor to initialize the database parameters.
        createSyntheticUtilityDatabase(outputFile)
            Create utility database and store it in the specified output file.
        createRandomNumbers(n, targetSum)
            Generate a list of random numbers with a specified target sum.
        save(outputFile)
            Save the generated utility database to a CSV file.
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
    Credits:
        The complete program was written by A.Hemanth sree sai under the supervision of Professor Rage Uday Kiran.
    """

    def __init__(self, totalTransactions: int, numOfItems: int, maxUtilRange: int, avgTransactionLength: int) -> None:
        """
        Constructor to initialize the database parameters.

        Parameters:
            totalTransactions (int): Number of transactions.
            numOfItems (int): Number of items.
            maxUtilRange (int): Maximum utility range.
            avgTransactionLength (int): The length of average transaction.
        """
        self.totalTransactions = totalTransactions
        self.numOfItems = numOfItems
        self.maxUtilRange = maxUtilRange
        self.avgTransactionLength = avgTransactionLength
        self.transactions = []
        self._startTime = float()
        self._endTime = float()
        self._memoryUSS = float()
        self._memoryRSS = float()

    def createSyntheticUtilityDatabase(self, outputFile: str) -> None:
        """
        Create utility database and store it in the specified output file.

        Parameters:
            outputFile (str): File name or path to store the database.
        """
        self._startTime = time.time()
        if self.avgTransactionLength > self.numOfItems:
            print("Error: avgTransactionLength cannot exceed numOfItems.")
            return

        with open(outputFile, 'w') as writer:
            for _ in range(self.totalTransactions):
                length = _rd.randint(1, self.avgTransactionLength + 20)
                items = [_rd.randint(1, self.numOfItems) for _ in range(length)]
                utilities = [_rd.randint(1, self.maxUtilRange) for _ in range(length)]

                # Generating 13 random numbers with a target sum of 2000
                randomNumbers = self.createRandomNumbers(13, 2000)

                # Checking if avgTransactionLength exceeds numOfItems
                if self.avgTransactionLength > self.numOfItems:
                    print("Error: avgTransactionLength cannot exceed numOfItems.")
                    return

                st = '\t'.join(map(str, items)) + '\t:' + str(sum(utilities)) + ':'
                st1 = '\t'.join(map(str, randomNumbers)) + '\t'

                writer.write(f"{st}{st1}\n")
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        self._endTime = time.time()

    def createRandomNumbers(self, n: int, targetSum: int) -> list[float]:
        """
        Generate a list of random numbers with a specified target sum.

        Parameters:
            n (int): Number of random numbers to generate.
            targetSum (int): Target sum for the generated random numbers.

        Returns:
            list: List of generated random numbers normalized and multiplied by the target sum.
        """
        randNumbers = [_rd.uniform(0, 1) for _ in range(n)]
        randSum = sum(randNumbers)
        normalizedNumbers = [num / randSum for num in randNumbers]
        result = [round(num * targetSum) for num in normalizedNumbers]
        return result

    def save(self, outputFile: str) -> None:
        """
        Save the generated utility database to a CSV file.

        Parameters:
            outputFile (str): File name or path to store the CSV file.
        """
        with open(outputFile, 'w') as f:
            for transaction in self.transactions:
                f.write('\t'.join(map(str, transaction)) + '\n')

    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

if __name__ == "__main__":
    obj = syntheticUtilityDatabase(100000, 870, 100, 10)
    obj.createSyntheticUtilityDatabase("T10_util-12.csv")
    print("create SyntheticUtilityDatabase is complete.")
    print("Total Memory in USS:", obj.getMemoryUSS())
    print("Total Memory in RSS", obj.getMemoryRSS())
    print("Total ExecutionTime in ms:", obj.getRuntime())
else:
    print("Error! The number of input parameters does not match the total number of parameters provided")



# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from PAMI.partialPeriodicPattern.pyspark import 4PGrowth as alg
#
#             obj = alg.parallel3PGrowth(iFile, minPS, period,numWorkers)
#
#             obj.mine()
#
#             partialPeriodicPatterns = obj.getPatterns()
#
#             print("Total number of partial periodic Patterns:", len(partialPeriodicPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDf()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#



__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.partialPeriodicPattern.pyspark import abstract as _ab
import validators as _validators
from urllib.request import urlopen as _urlopen
import sys as _sys
from pyspark import SparkContext, SparkConf
import pandas as pd
from deprecated import deprecated

_periodicSupport = float()
_period = float()
_lno = int()

class Node(object):
    """
    A class to represent the node of a tree

    :Attributes:

        item : int
            item of the node
        children : dict
            children of the node
        parent : class
            parent of the node
        tids : list.
            list of tids

    :Methods:

        _getTransactions()
            returns the list of transactions
        addChild(node)
            adds the child node to the parent node    
    
    """
    def __init__(self, item, children):
        """
        :param item : int
            item of the node
        :param children : dict
            children of the node    
        
        """
        self.item = item
        self.children = children 
        self.parent = None 
        self.tids =  []

    def _getTransactions(self):
        """
        returns the list of transactions

        """
        tids = self.tids
        for child in self.children.values():
            for t in child._getTransactions():
                t[0].append(child.item)
                yield t
        if (len(tids)>0):
            yield ([],tids)

    def addChild(self, node): 
        """
        adds the child node to the parent node

        :param node : class
                child node to be added
        """

        self.children[node.item] = node 
        node.parent = self



class Tree(object):
    """
    A class to represent the tree

    :Attributes:

        root : class
            root of the tree
        summaries : dict
            dictionary to store the summaries
        info : dict
            dictionary to store the information

    :Methods:

        add_transaction(transaction,tid)
            adds the transaction to the tree
        add_transaction_summ(transaction,tid_summ)
            adds the transaction to the tree
        get_condition_pattern(alpha)
            returns the condition pattern
        remove_node(node_val)
            removes the node from the tree
        get_ts(j)
            returns the ts
        getTransactions()
            returns the list of transactions
        merge(tree)
            merges the tree
        generate_patterns(prefix,glist,isResponsible = lambda x:True)
            generates the patterns    
    
    """

    def __init__(self):

        self.root = Node(None, {})
        self.summaries = {}
        self.info={}


    def add_transaction(self,transaction,tid):
        """
        adds the transaction to the tree

        :param transaction : list
                transaction to be added
        :param tid : int
                tid of the transaction

        :return: class returns the tree
        
        """
        curr_node=self.root
        for i in range(len(transaction)):
            if transaction[i] not in curr_node.children:
                new_node=Node(transaction[i],{})
                curr_node.addChild(new_node)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(new_node)
                else:
                    self.summaries[transaction[i]]=[new_node]                    
                curr_node=new_node                
            else:
                curr_node=curr_node.children[transaction[i]]            
        curr_node.tids.append(tid)
        return self
    
    def add_transaction_summ(self,transaction,tid_summ):
        """
        adds the transaction to the tree

        :param transaction : list
                transaction to be added
        :param tid_summ : list
                tid_summ of the transaction
        :return: class returns the tree
        
        """
        curr_node=self.root
        for i in range(len(transaction)):
            if transaction[i] not in curr_node.children:
                new_node=Node(transaction[i],{})
                curr_node.addChild(new_node)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(new_node)
                else:
                    self.summaries[transaction[i]]=[new_node]                    
                curr_node=new_node                
            else:
                curr_node=curr_node.children[transaction[i]]
        curr_node.tids+=tid_summ
        return self
        
    def get_condition_pattern(self,alpha):
        """
        returns the condition pattern

        :param alpha : int
                alpha value

        :return: list
                returns the list of patterns
        """
        final_patterns=[]
        final_sets=[]
        for i in self.summaries[alpha]:
            set1=i.tids
            set2=[]
            while(i.parent.item!=None):
                set2.append(i.parent.item)
                i=i.parent
            if(len(set2)>0):
                set2.reverse()
                final_patterns.append(set2)
                final_sets.append(set1)
        # print(final_patterns,final_sets)
        x,y,z=cond_trans(final_patterns,final_sets)#find this
        return x,y,z
    
    def remove_node(self,node_val):
        """
        removes the node from the tree

        :param node_val : int
                node value
        
        """
        for i in self.summaries[node_val]:
            i.parent.tids +=i.tids
            del i.parent.children[node_val]
            i=None

    def get_ts(self,j):
        """
        returns the ts
        :param j : int
                j value

        :return: list
                returns the list of ts
        """
        summari=[]
        for i in self.summaries[j]:
            summari+=i.tids
        return summari
    
    def getTransactions(self):
        """
        returns the list of transactions
        :return: list
                returns the list of transactions
        """
        return [x for x in self.root._getTransactions()]

    def merge(self,tree):
        """
        merges the tree

        :param tree : class
                tree to be merged

        :return: class
                returns the merged tree
        """
        for t in tree.getTransactions():
            t[0].reverse()
            self.add_transaction_summ(t[0], t[1])
        return self
  
    def generate_patterns(self,prefix,glist,isResponsible = lambda x:True):
        """
        generates the patterns

        :param prefix : list
                prefix of the pattern
        :param glist : list.
                list of items
        :param isResponsible : lambda function.
                lambda function to check the responsibility

        :return: list
                returns the list of patterns
        
        """
        for j in sorted(self.summaries,key= lambda x: (self.info.get(x),-x)):
            if(isResponsible(j)):
                rec_pattern=prefix.copy()
                rec_pattern.append(glist[j])
                yield (rec_pattern,self.info[j])
                patterns,tids,info=self.get_condition_pattern(j)
                conditional_tree=Tree()
                conditional_tree.info=info
                for pat in range(len(patterns)):
                    conditional_tree.add_transaction_summ(patterns[pat],tids[pat])
                if(len(patterns)>=1):
                    for li_m in conditional_tree.generate_patterns(rec_pattern,glist):
                        yield li_m
            self.remove_node(j)

class parallel3PGrowth(_ab._partialPeriodicPatterns):
    """
    :Description:   4PGrowth is fundamental approach to mine the partial periodic patterns in temporal database.

    :Reference:   Discovering Partial Periodic Itemsets in Temporal Databases,SSDBM '17: Proceedings of the 29th International Conference on Scientific and Statistical Database ManagementJune 2017
                  Article No.: 30 Pages 1–6https://doi.org/10.1145/3085504.3085535

    :param  iFile: str :
                   Name of the Input file to mine complete set of frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  period: float:
                   Minimum partial periodic...
    :param  periodicSupport: float:
                   Minimum partial periodic...

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        iFile : file
            Name of the Input file or path of the input file
        oFile : file
            Name of the output file or path of the output file
        periodicSupport: float or int or str
            The user can specify periodicSupport either in count or proportion of database size.
            If the program detects the data type of periodicSupport is integer, then it treats periodicSupport is expressed in count.
            Otherwise, it will be treated as float.
            Example: periodicSupport=10 will be treated as integer, while periodicSupport=10.0 will be treated as float
        period: float or int or str
            The user can specify period either in count or proportion of database size.
            If the program detects the data type of period is integer, then it treats period is expressed in count.
            Otherwise, it will be treated as float.
            Example: period=10 will be treated as integer, while period=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            it represents the total no of transactions
        tree : class
            it represents the Tree class
        finalPatterns : dict
            it represents to store the patterns

    :Methods:

        mine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets()
            Scans the dataset or dataframes and stores in list format
        partialPeriodicOneItem()
            Extracts the one-frequent patterns from transactions
        updateTransactions()
            updates the transactions by removing the aperiodic items and sort the transactions with items
            by decreasing support
        buildTree()
            constrcuts the main tree by setting the root node as null
        mine()
            main program to mine the partial periodic patterns

    **Executing the code on terminal:**
    ---------------------------------------
    .. code-block:: console


       Format:

       (.venv) $ python3 parallel3PGrowth.py <inputFile> <outputFile> <periodicSupport> <period>
    
       Examples:

       (.venv) $ python3 parallel3PGrowth.py sampleDB.txt patterns.txt 10.0 2.0

    **Sample run of the importing code:**
    -----------------------------------------
    .. code-block:: python

            from PAMI.partialPeriodicPattern.basic import 4PGrowth as alg

            obj = alg.4PGrowth(iFile, periodicSupport, period)

            obj.mine()

            partialPeriodicPatterns = obj.getPatterns()

            print("Total number of partial periodic Patterns:", len(partialPeriodicPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDf()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

                run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)


    **Credits:**
    -----------------
    The complete program was written by me under the supervision of Professor Rage Uday Kiran.\n

    """

    _periodicSupport = float()
    _period = float()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _rank = {}
    _rankdup = {}
    _lno = 0

    numPartitions = 5

    @deprecated("It is recommended to use mine() instead of mine() for mining process")
    def startMine(self):
        """
        Main method where the patterns are mined by constructing tree.
        """
        
        self.mine()

    def mine(self):
        """
        Main method where the patterns are mined by constructing tree.
        """

        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minPS is None:
            raise Exception("Please enter the Minimum Period-Support")

        self._period = self._convert(self._period)
        self._minPS = self._convert(self._minPS)
        minPS = self._minPS
        period = self._period

        APP_NAME = "4PGrowth"
        conf = SparkConf().setAppName(APP_NAME)
        sc = SparkContext(conf=conf).getOrCreate()

        self._startTime = _ab._time.time()

        data = sc.textFile(self._iFile, self.numPartitions).map(lambda x: [y for y in x.strip().split(self._sep)])
        # self.numPartitions = data.getNumPartitions()
        # numPartitions = 50
        freqItems, RecItems = self.getFrequentItems(data)
        # print(RecItems)

        trans = self.getFrequentItemsets(data, freqItems, self._period, self._minPS, dict(RecItems))
        a = trans.collect()

        # print(type(a))
        for k, v in a:
            string = "\t".join(k)
            # print(string,":",v)
            self._finalPatterns[string] = v

        # print(self._finalPatterns)
        #     print(k,":",v)
        # trans.saveAsTextFile('temp')
        self._endTime = _ab._time.time()
        sc.stop()

        # self._creatingItemSets()
        # generatedItems, pfList = self._partialPeriodicOneItem()
        # _minPS, _period, _lno = self._minPS, self._period, len(self._Database)
        # updatedTransactions = self._updateTransactions(generatedItems)
        # for x, y in self._rank.items():
        #     self._rankdup[y] = x
        # info = {self._rank[k]: v for k, v in generatedItems.items()}
        # Tree = self._buildTree(updatedTransactions, info)
        # patterns = Tree._generatePatterns([])
        # self._finalPatterns = {}
        # for i in patterns:
        #     s = self._savePeriodic(i[0])
        #     self._finalPatterns[s] = i[1]
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Partial Periodic Patterns were generated successfully using 4PGrowth algorithm ")

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._dbSize * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._dbSize * value)
            else:
                value = int(value)
        return value

    def cond_trans(self,cond_pat,cond_tids):
        """
        returns the condition pattern

        :param cond_pat : list
                condition pattern
        :param cond_tids : list
                condition tids

        :return: list
                returns the list of patterns
        """
        pat=[]
        tids=[]
        data1={}
        # print(cond_pat,cond_tids)
        for i in range(len(cond_pat)):
            for j in cond_pat[i]:
                if j in data1:
                    data1[j]=data1[j]+cond_tids[i]
                else:
                    data1[j]=cond_tids[i]

        up_dict={}
        for m in data1:
            up_dict[m]=getps(data1[m])
        up_dict={k: v for k,v in up_dict.items() if v>=minPS}
        count=0
        for p in cond_pat:
            p1=[v for v in p if v in up_dict]
            trans=sorted(p1, key= lambda x: (up_dict.get(x),-x), reverse=True)
            if(len(trans)>0):
                pat.append(trans)
                tids.append(cond_tids[count])
            count+=1
        return pat,tids,up_dict

    def getps(self,tid_list):
        """
        returns the periodic support

        :param tid_list : list.
                list of tids
        :return: int
                returns the periodic support
        """
        tid_list.sort()
        tids=tid_list
        cur=tids[0]
        pf=0
        for i in range(1,len(tids)):
            if tids[i]-cur<=self._period:
                pf+=1
            cur=tids[i]
        return pf


    def getPF(self,tid_list):
        """
        returns the periodic support

        :param tid_list : list.
                list of tids

        :return: int
                returns the periodic support
        """
        tid_list.sort()
        tids=tid_list
        cur=tids[0]
        pf=0
        for i in range(1,len(tids)):
            if tids[i]-cur<=self._period:
                pf+=1
            cur=tids[i]
        return pf

    def getFrequentItems(self,data):
        """
        returns the frequent items

        :param data : list
                list of transactions

        :return: list
                returns the list of frequent items
        """
        
        # t1 = time.time()
        singleItems = data.flatMap(lambda x: [(y,[int(x[0])]) for y in x[1:]])
        RecItems=singleItems.reduceByKey(lambda x,y: x + y)\
        .map(lambda c :(c[0],self.getPF(c[1]))).filter(lambda c: c[1]>=self._minPS).collect()
        # RecItems=itemsWtTids.filter(lambda c : getPF(c[1])>=minPS)
        RecItemSorted=[x for (x,y) in sorted(RecItems,key=lambda x : -x[1])]
        # t2 = time.time()
        # print ("size one", t2-t1)
        return RecItemSorted, RecItems

    def getFrequentItemsets(self,data,perFreqItems,per,minPS, PSinfo):
        """
        returns the frequent itemsets

        :param data : list.
                list of transactions
        :param perFreqItems : list.
                list of frequent items
        :param per : int
                period
        :param minPS : int
                minimum periodic support
        :param PSinfo : dict
                dictionary to store the information

        :return: list
                returns the list of frequent itemsets

        """
        # t1 = time.time()
        rank = dict([(index, item) for (item,index) in enumerate(perFreqItems)]) 
        numPartitions = data.getNumPartitions()
        workByPartition = data.flatMap(lambda basket:self.genCondTransactions(basket[0], basket[1:],rank,numPartitions))
        # emptyTree =ppp.Tree()
        c=0
        inf={}
        # print(rank)
        # print(PSinfo)
        v=len(perFreqItems)
        for i in rank:
            inf[rank[i]]=PSinfo[i]
            c+=1
        # print(inf)
        emptyTree = Tree()

        emptyTree.info=inf    
        forest = workByPartition.aggregateByKey(emptyTree,lambda tree,transaction: tree.add_transaction(transaction[1:], transaction[0]),lambda tree1,tree2: tree1.merge(tree2))
        itemsets = forest.flatMap(lambda partId_bonsai: partId_bonsai[1].generate_patterns([],perFreqItems, lambda x: self.getPartitionId(x,numPartitions) == partId_bonsai[0]))
        # t2 = time.time()
        # print ("main", t2-t1)
        return itemsets
    
    def genCondTransactions(self,tid,basket, rank, nPartitions):
        """
        returns the conditional transactions

        :param tid : int
                tid of the transaction
        :param basket : list.
                list of items
        :param rank : dict
                dictionary to store the rank
        :param nPartitions : int
                number of partitions

        :return: list
                returns the list of conditional transactions
        
        """
        #translate into new id's using rank
        filtered = [rank[x] for x in basket if x in rank.keys()]
        #sort basket in ascending rank
        filtered = sorted(filtered)
        output = {}
        pc=0
        for i in range(len(filtered)-1, -1, -1):
            item = filtered[i]
            partition = self.getPartitionId(item, nPartitions)
            if partition not in output.keys():
                output[partition] = [int(tid)]+filtered[:i+1]
                pc+=1
                if pc==nPartitions:
                    break
        return [x for x in output.items()]

    def getPartitionId(self,key, nPartitions):
        return key % nPartitions
    
    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            # print(a,b)
            data.append([a.replace('\t', ' '), b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'minPS'])
        return dataFrame

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        print("Total number of Partial Periodic Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())

    def setPartitions(self,nums):
        self.numPartitions = nums

def cond_trans(cond_pat,cond_tids):
    """
    returns the condition pattern

    :param cond_pat : list
            condition pattern
    :param cond_tids : list
            condition tids

    """
    
    pat=[]
    tids=[]
    data1={}
    # print(cond_pat,cond_tids)
    for i in range(len(cond_pat)):
        for j in cond_pat[i]:
            if j in data1:
                data1[j]=data1[j]+cond_tids[i]
            else:
                data1[j]=cond_tids[i]

    up_dict={}
    for m in data1:
        up_dict[m]=getps(data1[m])
    up_dict={k: v for k,v in up_dict.items() if v>=minPS}
    count=0
    for p in cond_pat:
        p1=[v for v in p if v in up_dict]
        trans=sorted(p1, key= lambda x: (up_dict.get(x),-x), reverse=True)
        if(len(trans)>0):
            pat.append(trans)
            tids.append(cond_tids[count])
        count+=1
    return pat,tids,up_dict

def getps(tid_list):
    """
    
    returns the periodic support

    :param tid_list : list.
            list of tids

    """
    tid_list.sort()
    tids=tid_list
    cur=tids[0]
    pf=0
    for i in range(1,len(tids)):
        if tids[i]-cur<=period:
            pf+=1
        cur=tids[i]
    return pf


def getPF(self,tid_list):
    tid_list.sort()
    tids=tid_list
    cur=tids[0]
    pf=0
    for i in range(1,len(tids)):
        if tids[i]-cur<=period:
            pf+=1
        cur=tids[i]
    return pf
    
if __name__ == "__main__":
    _ap = str()
    if len(_sys.argv) == 5 or len(_sys.argv) == 6:
        period = 0
        minPS = 0
        if len(_sys.argv) == 6:
            _ap = parallel3PGrowth(_sys.argv[1], _sys.argv[3], _sys.argv[4], _sys.argv[5])
        if len(_sys.argv) == 5:
            _ap = parallel3PGrowth(_sys.argv[1], _sys.argv[3], _sys.argv[4])
        _ap.mine()
        print("Total number of Partial Periodic Patterns:", len(_ap.getPatterns()))
        _ap.save(_sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:",  _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")
        minPS = 500
        period = 50000


        _ap = parallel3PGrowth('Temporal_T10I4D100K.csv', minPS, period, '\t')
        _ap.setPartitions(20)
        _ap.mine()
        


# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from PAMI.contiguousFrequentPattern.basic import PositionMining as alg
#
#             obj =alg.PositionMining(minsup=5,data="Data.csv")
#
#             obj.mine()
#
#             Patterns = obj.getPatterns()
#
#             print("Total number of high utility frequent Patterns:", len(Patterns))
#
#             obj.save("output")
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#





import pandas as pd
import numpy as np
import math
from PAMI.contiguousFrequentPattern import abstract as _ab
from deprecated import deprecated


class Node:
    def __init__(self,symbol,leaf=False):
        self._val=symbol
        self.children=[None for i in range(26)]
        self.leaf=leaf
        self.freq={}
        self.count=1
        self.seq=""




class PositionMining:
    """
    About this algorithm
    ====================

    :Description:  describe the algorithm

    :Reference: provide the reference of the algorithm with URL of the paper, if possible

    :param  parameter: parameterType :
                    description of the parameters. Parameters are the variables used in problem definition of the model.

    :Attributes:

        min_IG: float
                minimum threshold for information gain
            
        min_conf: float
                minimum threshold for confidence

        datapath: .csv file consisting of two id,seq fields respectively in order
    

    Credits
    =======
             The complete program was written by Shiridi kumar under the supervision of Professor uday rage.

    """

    def __init__(self,minsup,datapath,maxlength=20):
        self.min_sup=minsup
        self.datapath=datapath
        self.maxlength=maxlength
    

    def readData(self):
        df=pd.read_csv(self.datapath)
        vals=df.values
        self.seq_prefixes={}
        # prev=0
        # self.seq_prefixes[vals[0][0]]=len(vals[0])
        # for i in range(1,len(vals)):
        #     self.seq_prefixes[vals[i]]
        self.data=vals
        # print(self.data)

    def getfreqs(self):
        """
        Initial scan of database where frequent length one candidate will be mined
        """
        self.symbol_freq={"A":set(),"G":set(),"C":set(),"T":set()}
        self.total_length=0
        curr_pos=0
        # print(self.data)
        for i in range(len(self.data)):
            seq=self.data[i][1]
            for j in range(len(seq)):
                self.symbol_freq[seq[j]].add(curr_pos)
                curr_pos+=1
            curr_pos+=1
            self.total_length+=len(seq)

        temp={}
        for i in self.symbol_freq:
            if(len(self.symbol_freq[i])>=self.min_sup):
                temp.update({i:self.symbol_freq[i]})
        self.symbol_freq=temp

    
    

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """

        return self.frequentPatterns


    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        seqs=[]
        sup=[]
        for i in self.frequentPatterns:
            seqs.append(i)
            sup.append(self.frequentPatterns[i])


        dataFrame =pd.DataFrame()
        dataFrame["Patterns"]=seqs
        dataFrame["Support"]=sup
        return dataFrame


    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        df=self.getPatternsAsDataFrame()
        df.to_csv(outFile)


        
    def get_Klength_patterns(self,k):
        """
        Get frequent patterns of klength

        :param k : length of the pattern
        :type k: dictionary of frequent patterns
        """

        dic={i:len(self.table[k][i]) for i in self.table[k]}
        return dic
    


    def getPattern_positions(self,pattern):
        length=len(pattern)
        positions=self.table[length][pattern]
        return positions

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self._endTime-self._startTime
    
    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of High Utility Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in seconds:", self.getRuntime())


    def join(self,db,length):
        """
        Generating l+1 frequent patterns using two l length frequent patterns

        :param db:current l length frequent patterns table consisting of their positions
        :type db: HashTable
        :param length:current length of the frequent candidates generated
        :type length: positive integer
        """
        for seq1 in db:
            for seq2 in db:
                if(seq1!=seq2):
                    if(length==1):
                        word=seq1+seq2
                        # print(seq1,seq2,db[seq1],db[seq2])
                        minus_1={i-1 for i in db[seq2]}
                        positions=db[seq1].intersection(minus_1)
                        if(len(positions)>=self.min_sup):
                            self.table[length+1].update({word:positions})


                    else:
                        if(seq1[1:]== seq2[:-1]):
                            word=seq1+seq2[-1]
                            minus_1={i-1 for i in db[seq2]}
                            positions=db[seq1].intersection(minus_1)
                            if(len(positions)>=self.min_sup):
                                self.table[length+1].update({word:positions})
        

    def mineNext_candidates(self):
        """
        Mining frequent patterns along with their positions from length 1 frequent candidates
        """
        while self.current_candidate<self.maxlength-1:
            curr=self.table[self.current_candidate]
            self.join(curr,self.current_candidate)
            self.current_candidate+=1

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Pattern mining process will start from here
        """
        # pass
        self.mine()

    def mine(self):
        """
        Pattern mining process will start from here
        """
        # pass
        self._startTime = _ab._time.time()
        self.table = {i: {} for i in range(1, self.maxlength)}
        self.readData()

        self.getfreqs()
        temp = self.symbol_freq
        self.table.update({1: temp})
        self.current_candidate = 1
        self.mineNext_candidates()
        self.frequentPatterns = {}
        for length in self.table:
            temp = self.table[length]
            for pattern in temp:
                self.frequentPatterns.update({pattern: len(temp[pattern])})

        process = _ab._psutil.Process(_ab._os.getpid())
        self._endTime = _ab._time.time()
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss


# csvToBitInteger is a code used to convert the csv into Bit Integer.
#
# **Importing this algorithm into a python program**
#  --------------------------------------------------------
#
#             from PAMI.extras.csvParquet import csvToBitInteger as cp
#
#             obj = cp.csvToBitInteger(iFile, "\t ")
#
#             obj.save()
#
#             obj.csvToBitInteger("FileName") # To generate file in form of BitInteger.
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
import pandas as pd

class csvToBitInteger():
    """

    :Description:   csvToBitInteger is a code used to convert the csv into Bit Integer

    :param  iFile: str :
            Name of the Input file
    :param  sep: str :
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.



    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.csvParquet import csvToBitInteger as cp

            obj = cp.csvToBitInteger(File, "\t")

            obj.save()

            obj.csvToBitInteger("FileName") # To generate file in form of BitInteger.

    """

    def __init__(self, iFile: str, sep: str = '\t'):
        self._iFile = iFile
        self._sep = sep


def csvToBitInteger(file, output, sep = "\t"):

    fileData = {}
    rename = {}
    counter = 0
    maxTS = 0


    with open(file,'r') as f:
        
        for line in f:
            line = line.strip().split(sep)
            timestamp = int(line[0])
            if timestamp > maxTS:
                maxTS = timestamp

            line = line[1:]

            for item in line:
                if item not in rename:
                    rename[item] = counter
                    counter += 1
                if rename[item] not in fileData:
                    fileData[rename[item]] = [timestamp]
                else:
                    fileData[rename[item]].append(timestamp)

    fileData = dict(sorted(fileData.items(), key = lambda x: len(x[1])))


    newRep = {}
    # arraySize = maxTS // 32 + 1

    # print("Filedata:",len(fileData))
    for k,v in fileData.items():
        bitRep = np.zeros(maxTS + 1, dtype=np.uint32)
        for i in range(len(v)):
            bitRep[v[i]] = 1
        newRep[k] = bitRep
        # print(k,":",len(v))

    # print("Creating dataframe...")
    df = pd.DataFrame(newRep)
    # print("Transposing dataframe...")
    df = df.T

    cols = []

    for i in range(len(df.columns)):
        cols.append(str(i))

    df.columns = cols

    # print("Writing to parquet...")
    df.to_parquet(output)




# FPGrowth is one of the fundamental algorithm to discover frequent patterns in a transactional database. It stores the database in compressed fp-tree decreasing the memory usage and extracts the patterns from tree.It  employs downward closure property to  reduce the search space effectively.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.frequentPattern.basic import FPGrowth as alg
#
#             iFile = 'sampleDB.txt'
#
#             minSup = 10  # can also be specified between 0 and 1
#
#             obj = alg.FPGrowth(iFile, minSup)
#
#             obj.mine()
#
#             frequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PAMI.frequentPattern.basic import abstract as _fp
from typing import List, Dict, Tuple, Any
from deprecated import deprecated
from itertools import combinations
from collections import Counter

_minSup = str()
_fp._sys.setrecursionlimit(20000)


class _Node:
    """
    A class used to represent the node of frequentPatternTree

    :**Attributes**:    - **itemId** (*int*) -- *storing item of a node.*
                        - **counter** (*int*) -- *To maintain the support of node.*
                        - **parent** (*node*) -- *To maintain the parent of node.*
                        - **children** (*list*) -- *To maintain the children of node.*

    :**Methods**:   - **addChild(node)** -- *Updates the nodes children list and parent for the given node.*
    """

    def __init__(self, item, count, parent) -> None:
        self.item = item
        self.count = count
        self.parent = parent
        self.children = {}

    def addChild(self, item, count = 1) -> Any:
        """

        Adds a child node to the current node with the specified item and count.

        :param item: The item associated with the child node.
        :type item: List
        :param count: The count or support of the item. Default is 1.
        :type count: int
        :return: The child node added.
        :rtype: List
        """
        if item not in self.children:
            self.children[item] = _Node(item, count, self)
        else:
            self.children[item].count += count
        return self.children[item]
    
    def traverse(self) -> Tuple[List[int], int]:
        """
        Traversing the tree to get the transaction

        :return: transaction and count of each item in transaction
        :rtype: Tuple, List and int
        """
        transaction = []
        count = self.count
        node = self.parent
        while node.parent is not None:
            transaction.append(node.item)
            node = node.parent
        return transaction[::-1], count


class FPGrowth(_fp._frequentPatterns):
    """
    **About this algorithm**

    :**Description**:   FPGrowth is one of the fundamental algorithm to discover frequent patterns in a transactional database. It stores the database in compressed fp-tree decreasing the memory usage and extracts the patterns from tree.It employs downward closure property to  reduce the search space effectively.

    :**Reference**:  Han, J., Pei, J., Yin, Y. et al. Mining Frequent Patterns without Candidate Generation: A Frequent-Pattern
                     Tree Approach. Data  Mining and Knowledge Discovery 8, 53–87 (2004). https://doi.org/10.1023

    :**Parameters**:    - **iFile** (*str or URL or dataFrame*) -- *Name of the Input file to mine complete set of frequent patterns.*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of frequent patterns.*
                        - **minSup** (*int or float or str*) -- *The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count. Otherwise, it will be treated as float.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:    - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **finalPatterns** (*dict*) -- *Storing the complete set of patterns in a dictionary variable.*
                        - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **Database** (*list*) -- *To store the transactions of a database in list.*
                        - **mapSupport** (*Dictionary*) -- *To maintain the information of item and their frequency.*
                        - **tree** (*class*) --  *it represents the Tree class.*


    **Execution methods**

    **Terminal command**

    .. code-block:: console

      Format:

      (.venv) $ python3 FPGrowth.py <inputFile> <outputFile> <minSup>

      Example Usage:

      (.venv) $ python3 FPGrowth.py sampleDB.txt patterns.txt 10.0

    .. note:: minSup can be specified  in support count or a value between 0 and 1.


    **Calling from a python program**

    .. code-block:: python

            from PAMI.frequentPattern.basic import FPGrowth as alg

            iFile = 'sampleDB.txt'

            minSup = 10  # can also be specified between 0 and 1

            obj = alg.FPGrowth(iFile, minSup)

            obj.mine()

            frequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.savePatterns(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    **Credits:**

    The complete program was written by P. Likhitha and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """

    __startTime = float()
    __endTime = float()
    _minSup = str()
    __finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    __memoryUSS = float()
    __memoryRSS = float()
    __Database = []
    __mapSupport = {}
    __lno = 0
    __rank = {}
    __rankDup = {}

    def __init__(self, iFile, minSup, sep='\t') -> None:
        super().__init__(iFile, minSup, sep)

    def __creatingItemSets(self) -> None:
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self.__Database = []
        if isinstance(self._iFile, _fp._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.__Database = self._iFile['Transactions'].tolist()
                self.__Database = [x.split(self._sep) for x in self.__Database]
            else:
                print("The column name should be Transactions and each line should be separated by tab space or a seperator specified by the user")
                

            #print(self.Database)
        if isinstance(self._iFile, str):
            if _fp._validators.url(self._iFile):
                data = _fp._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self.__Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self.__Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def __convert(self, value) -> float:
        """

        To convert the type of user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        :rtype: float
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.__Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.__Database) * value)
            else:
                value = int(value)
        return value
    
    def _construct(self, items, data, minSup):
        """
        Constructs the FP-tree from the given transactions.

        :param items: A dictionary containing item frequencies.
        :type items: Dict
        :param data: A list of transactions.
        :type data: List
        :param minSup: The minimum support threshold.
        :type minSup: int
        :return: The root node of the constructed FP-tree and a dictionary containing information about nodes associated with each item.
        :rtype: Tuple[_Node, Dict]
        """

        items = {k: v for k, v in items.items() if v >= minSup}

        root = _Node([], 0, None)
        itemNodes = {}
        for line in data:
            currNode = root
            line = sorted([item for item in line if item in items], key = lambda x: items[x], reverse = True)
            for item in line:
                currNode = currNode.addChild(item)
                if item in itemNodes:
                    itemNodes[item][0].add(currNode)
                    itemNodes[item][1] += 1
                else:
                    itemNodes[item] = [set([currNode]), 1]

        return root, itemNodes

    def _all_combinations(self, arr):
        """

        Generates all possible combinations of items from a given transaction.

        :param arr: A list of items in a transaction.
        :type arr: List
        :return: A list containing all possible combinations of items.
        :rtype: List
        """

        all_combinations_list = []
        for r in range(1, len(arr) + 1):
            all_combinations_list.extend(combinations(arr, r))
        return all_combinations_list
    
    def _recursive(self, root, itemNode, minSup, patterns):
        """

         Recursively explores the FP-tree to generate frequent patterns.

         :param root: The root node of the current subtree.
         :type root: _Node
         :param itemNode: A dictionary containing information about the nodes associated with each item.
         :type itemNode: Dict
         :param minSup: The minimum support threshold.
         :type minSup: int
         :param patterns: A dictionary to store the generated frequent patterns.
         :type patterns: Dict
        """
        itemNode = {k: v for k, v in sorted(itemNode.items(), key = lambda x: x[1][1])}

        for item in itemNode:
            if itemNode[item][1] < self._minSup:
                break 

            newRoot = _Node(root.item + [item], 0, None)
            # pat = "\t".join([str(i) for i in newRoot.item])
            # self.__finalPatterns[pat] = itemNode[item][1]
            self._finalPatterns[tuple(newRoot.item)] = itemNode[item][1]
            newItemNode = {}

            if len(itemNode[item][0]) == 1:
                transaction, count = itemNode[item][0].pop().traverse()
                if len(transaction) == 0:
                    continue
                combination = self._all_combinations(transaction)
                for comb in combination:
                    # pat = "\t".join([str(i) for i in comb])
                    # pat = pat + "\t" + "\t".join([str(i) for i in newRoot.item])
                    # self.__finalPatterns[pat] = count
                    self._finalPatterns[tuple(list(comb) + newRoot.item)] = count
                pass


            itemCount = {}
            transactions = {}
            for node in itemNode[item][0]:
                transaction, count = node.traverse()
                if len(transaction) == 0:
                    continue
                if tuple(transaction) in transactions:
                    transactions[tuple(transaction)] += count
                else:
                    transactions[tuple(transaction)] = count


                for transaction_item in transaction:
                    if transaction_item in itemCount:
                        itemCount[transaction_item] += count
                    else:
                        itemCount[transaction_item] = count


            # remove items that are below minSup
            itemCount = {k: v for k, v in itemCount.items() if v >= minSup}
            if len(itemCount) == 0:
                continue

            for transaction, count in transactions.items():
                transaction = sorted([item for item in transaction if item in itemCount], key = lambda x: itemCount[x], reverse = True)
                currNode = newRoot
                for item_ in transaction:
                    currNode = currNode.addChild(item_, count)
                    if item_ in newItemNode:
                        newItemNode[item_][0].add(currNode)
                        newItemNode[item_][1] += count
                    else:
                        newItemNode[item_] = [set([currNode]), count]

            if len(newItemNode) < 1:
                continue

            # mine(newRoot, newItemNode, minSup, patterns)
            self._recursive(newRoot, newItemNode, minSup, patterns)


    def mine(self) -> None:
        """
        Main program to start the operation
        """
        global _minSup
        self.__startTime = _fp._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minSup is None:
            raise Exception("Please enter the Minimum Support")
        self.__creatingItemSets()
        self._minSup = self.__convert(self._minSup)
        _minSup = self._minSup

        itemCount = Counter()
        for line in self.__Database:
            itemCount.update(line)

        root, itemNode = self._construct(itemCount, self.__Database, self._minSup)
        self._recursive(root, itemNode, self._minSup, self.__finalPatterns)
        
        print("Frequent patterns were generated successfully using frequentPatternGrowth algorithm")
        self.__endTime = _fp._time.time()
        self.__memoryUSS = float()
        self.__memoryRSS = float()
        process = _fp._psutil.Process(_fp._os.getpid())
        self.__memoryUSS = process.memory_full_info().uss
        self.__memoryRSS = process.memory_info().rss

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Starting the mining process
        """
        self.mine()

    def getMemoryUSS(self) -> float:
        """

        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.__memoryUSS

    def getMemoryRSS(self) -> float:
        """

        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.__memoryRSS

    def getRuntime(self) -> float:
        """

        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.__endTime - self.__startTime
    

    def getPatternsAsDataFrame(self) -> _fp._pd.DataFrame:
        """

        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        # # dataframe = {}
        # # data = []
        # # for a, b in self.__finalPatterns.items():
        # #     data.append([a.replace('\t', ' '), b])
        # #     dataframe = _fp._pd.DataFrame(data, columns=['Patterns', 'Support'])
        # dataFrame = _fp._pd.DataFrame(list(self._finalPatterns.items()), columns=['Patterns', 'Support'])
        dataFrame = _fp._pd.DataFrame(list([[" ".join(x), y] for x,y in self._finalPatterns.items()]), columns=['Patterns', 'Support'])

        return dataFrame

    def save(self, outFile: str, seperator = "\t" ) -> None:
        """

        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csvfile
        :param seperator: variable to store the separator
        :type seperator: string
        :return: None
        """
        with open(outFile, 'w') as f:
            for x, y in self._finalPatterns.items():
                x = seperator.join(x)
                f.write(f"{x}:{y}\n")

    def getPatterns(self) -> Dict[str, int]:
        """

        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns
    
    def printResults(self) -> None:
        """
        This function is used to print the results
        """
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_fp._sys.argv) == 4 or len(_fp._sys.argv) == 5:
        if len(_fp._sys.argv) == 5:
            _ap = FPGrowth(_fp._sys.argv[1], _fp._sys.argv[3], _fp._sys.argv[4])
        if len(_fp._sys.argv) == 4:
            _ap = FPGrowth(_fp._sys.argv[1], _fp._sys.argv[3])
        _ap.mine()
        _ap.mine()
        print("Total number of Frequent Patterns:", len( _ap.getPatterns()))
        _ap.save(_fp._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



from vertex import Vertex
from edge import Edge
from dfsCode import DFSCode

class UncertainGraph:
    def __init__(self, id, vertexMap=None, dfsCode=None):
        self.id = id
        self.vertexMap = {}
        self.implicatedGraphs = []

        if vertexMap is not None:
            self.vertexMap = vertexMap

        if dfsCode is not None:
            self.buildFromDFSCode(dfsCode)

    def buildFromDFSCode(self, dfsCode):
        for ee in dfsCode.getEeList():
            v1, v2, v1Label, v2Label, eLabel = ee.getV1(), ee.getV2(), ee.getVLabel1(), ee.getVLabel2(), ee.getEdgeLabel()

            e = Edge(v1, v2, eLabel, 1.0)  # Assuming existenceProbability is 1.0 for all edges in DFS code
            if v1 not in self.vertexMap:
                self.vertexMap[v1] = Vertex(v1, v1Label)
            if v2 not in self.vertexMap:
                self.vertexMap[v2] = Vertex(v2, v2Label)

            self.vertexMap[v1].addEdge(e)
            self.vertexMap[v2].addEdge(e)

    def getVertexMap(self):
        return self.vertexMap

    def toDFSCode(self):
        dfsCode = DFSCode()
        for vertex in self.vertexMap.values():
            for edge in vertex.getEdgeList():
                v1, v2 = edge.v1, edge.v2
                if v1 < v2:
                    v1Label, v2Label = self.vertexMap[v1].getLabel(), self.vertexMap[v2].getLabel()
                else:
                    v1, v2 = v2, v1
                    v1Label, v2Label = self.vertexMap[v1].getLabel(), self.vertexMap[v2].getLabel()
                eLabel = edge.edgeLabel
                dfsCode.addEdge(v1, v2, v1Label, v2Label, eLabel)
        return dfsCode
